import os
import sys

from rich.console import Console


def run(**kwargs):
    'Profile data source, run assertions, and generate report(s). By default, the raw results and reports are saved in ".piperider/outputs".'

    from piperider_cli.cli_utils import DbtUtil
    from piperider_cli.cli_utils.cloud import CloudConnectorHelper
    from piperider_cli.configuration import FileSystem, is_piperider_workspace_exist
    from piperider_cli.error import DbtProjectNotFoundError, PipeRiderConflictOptionsError
    from piperider_cli.exitcode import EC_ERR_TEST_FAILED, EC_WARN_NO_PROFILED_MODULES
    from piperider_cli.generate_report import GenerateReport
    from piperider_cli.guide import Guide
    from piperider_cli.initializer import Initializer
    from piperider_cli.runner import Runner

    datasource = kwargs.get('datasource')
    table = kwargs.get('table')
    output = kwargs.get('output')
    open_report = kwargs.get('open')
    enable_share = kwargs.get('share')
    skip_report = kwargs.get('skip_report')
    dbt_target_path = kwargs.get('dbt_target_path')
    dbt_list = kwargs.get('dbt_list')
    force_upload = kwargs.get('upload')
    project_name = kwargs.get('project')
    select = kwargs.get('select')
    state = kwargs.get('state')

    if project_name is not None:
        os.environ.get('PIPERIDER_API_PROJECT')

    console = Console()
    env_dbt_resources = os.environ.get('PIPERIDER_DBT_RESOURCES')

    # True -> 1, False -> 0
    if sum([True if table else False, dbt_list, env_dbt_resources is not None]) > 1:
        console.print("[bold red]Error:[/bold red] "
                      "['--table', '--dbt-list'] are mutually exclusive")
        sys.exit(1)

    # Search dbt project config files
    dbt_project_dir = kwargs.get('dbt_project_dir')
    no_auto_search = kwargs.get('no_auto_search')
    dbt_project_path = DbtUtil.get_dbt_project_path(dbt_project_dir, no_auto_search, recursive=False)
    dbt_profiles_dir = kwargs.get('dbt_profiles_dir')
    if dbt_project_path:
        working_dir = os.path.dirname(dbt_project_path) if dbt_project_path.endswith('.yml') else dbt_project_path
        FileSystem.set_working_directory(working_dir)
        if dbt_profiles_dir:
            FileSystem.set_dbt_profiles_dir(dbt_profiles_dir)
        # Only run initializer when dbt project path is provided
        Initializer.exec(dbt_project_path=dbt_project_path, dbt_profiles_dir=dbt_profiles_dir, interactive=False)
    elif is_piperider_workspace_exist() is False:
        raise DbtProjectNotFoundError()

    dbt_resources = None
    if select and dbt_list is True:
        raise PipeRiderConflictOptionsError(
            'Cannot use options "--select" with "--dbt-list"',
            hint='Remove "--select" option and use "--dbt-list" instead.'
        )

    if dbt_list:
        dbt_resources = DbtUtil.read_dbt_resources(sys.stdin)
    if env_dbt_resources is not None:
        dbt_resources = DbtUtil.read_dbt_resources(env_dbt_resources)

    ret = Runner.exec(datasource=datasource,
                      table=table,
                      output=output,
                      skip_report=skip_report,
                      dbt_target_path=dbt_target_path,
                      dbt_resources=dbt_resources,
                      dbt_select=select,
                      dbt_state=state,
                      report_dir=kwargs.get('report_dir'),
                      skip_datasource_connection=kwargs.get('skip_datasource'))
    if ret in (0, EC_ERR_TEST_FAILED, EC_WARN_NO_PROFILED_MODULES):
        if enable_share:
            force_upload = True

        auto_upload = CloudConnectorHelper.is_auto_upload()
        is_cloud_view = (force_upload or auto_upload)

        if not skip_report:
            GenerateReport.exec(None, kwargs.get('report_dir'), output, open_report, is_cloud_view)

        if ret == EC_WARN_NO_PROFILED_MODULES:
            # No module was profiled
            if dbt_list or dbt_resources or select:
                Guide().show('No resources was profiled. Please check given "--select", "--dbt-list" option or '
                             'environment variable "PIPERIDER_DBT_RESOURCES" to choose the resources to profile.')
                ret = 0

        if CloudConnectorHelper.is_login() and is_cloud_view:
            ret = CloudConnectorHelper.upload_latest_report(report_dir=kwargs.get('report_dir'),
                                                            debug=kwargs.get('debug'),
                                                            open_report=open_report, enable_share=enable_share,
                                                            project_name=project_name)
        elif not CloudConnectorHelper.is_login() and is_cloud_view:
            console = Console()
            console.print('[bold yellow]Warning: [/bold yellow]The report is not uploaded due to not logged in.')

    if ret != 0:
        if ret != EC_WARN_NO_PROFILED_MODULES:
            sys.exit(ret)
    return ret



import json
import logging
import os
import re
import sys
import webbrowser
from datetime import datetime

from dateutil import tz
from rich.console import Console
from ruamel import yaml

PIPERIDER_USER_HOME = os.path.expanduser('~/.piperider')
PIPERIDER_USER_PROFILE = os.path.join(PIPERIDER_USER_HOME, 'profile.yml')


def create_logger(name) -> logging.Logger:
    log_level = logging.WARNING
    if os.environ.get('PIPERIDER_LOG_LEVEL') == 'DEBUG':
        log_level = logging.DEBUG
    if os.environ.get('PIPERIDER_LOG_LEVEL') == 'INFO':
        log_level = logging.INFO

    log = logging.getLogger(name)
    log.setLevel(log_level)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(log_level)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    log.addHandler(handler)
    return log


def is_executed_manually() -> bool:
    # some CI service puts the "CI" var
    if os.environ.get('CI', 'false') == 'true':
        return False

    # For CiecleCI exceptions
    if os.environ.get('CIRCLECI', 'false') == 'true':
        return False

    # if not tty, it probably runs automatically
    if not sys.stdout.isatty():
        return False

    return True


def get_assertion_dir():
    return os.environ.get('PIPERIDER_CUSTOM_ASSERTIONS_PATH', '')


def set_assertion_dir(dir):
    snap_dir = os.path.join(os.path.normpath(dir), '__snapshots__')
    if not os.path.exists(snap_dir):
        os.mkdir(snap_dir)
    os.environ['PIPERIDER_CUSTOM_ASSERTIONS_PATH'] = snap_dir


def get_version():
    version_file = os.path.normpath(os.path.join(os.path.dirname(__file__), 'VERSION'))
    with open(version_file) as fh:
        version = fh.read().strip()
        return version


def get_sentry_dns():
    dns_file = os.path.normpath(os.path.join(os.path.dirname(__file__), 'SENTRY_DNS'))
    with open(dns_file) as f:
        dns = f.read().strip()
        return dns


def get_user_id():
    with open(PIPERIDER_USER_PROFILE, 'r') as f:
        user_profile = yaml.YAML().load(f)
        return user_profile.get('user_id')


def clone_directory(src, dst):
    if sys.version_info >= (3, 8):
        # dirs_exist_ok only available after 3.8
        import shutil
        shutil.copytree(src, dst, dirs_exist_ok=True)
    else:
        from distutils.dir_util import copy_tree
        copy_tree(src, dst)


def datetime_to_str(input, to_tzlocal=False):
    if not isinstance(input, datetime):
        raise ValueError(f"Invalid type: expect 'datetime', got '{type(input).__name__}'")
    if to_tzlocal:
        input = convert_to_tzlocal(input)
    output = input.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
    return output


def str_to_datetime(input: str, to_tzlocal=False):
    output = datetime.strptime(input, '%Y-%m-%dT%H:%M:%S.%fZ')
    if to_tzlocal:
        output = convert_to_tzlocal(output)
    return output


def convert_to_tzlocal(input):
    if not isinstance(input, datetime):
        raise ValueError(f"Invalid type: expect 'datetime', got '{type(input).__name__}'")
    from_zone = tz.tzutc()
    to_zone = tz.tzlocal()
    return input.replace(tzinfo=from_zone).astimezone(to_zone)


__version__ = get_version()


def set_sentry_env():
    if '.dev' in __version__:
        return 'development'
    elif re.match(r'^\d+\.\d+\.\d+\.\d{8}[a|b|rc]?.*$', __version__):
        return 'nightly'
    elif 'a' in __version__:
        return 'alpha'
    elif 'b' in __version__:
        return 'beta'
    elif 'rc' in __version__:
        return 'release-candidate'
    return 'production'


sentry_dns = get_sentry_dns()
sentry_env = set_sentry_env()


def ensure_directory_writable(directory):
    d = os.path.abspath(directory)

    if os.path.exists(d):
        if not os.path.isdir(directory):
            return False
        return os.access(d, os.W_OK)
    else:
        try:
            os.makedirs(directory)
            return True
        except BaseException:
            return False


def raise_exception_when_directory_not_writable(output):
    if output:
        if not ensure_directory_writable(output):
            raise Exception(f'The path "{output}" is not writable')


def safe_load_yaml(file_path):
    try:
        with open(file_path, 'r') as f:
            payload = yaml.safe_load(f)
    except yaml.YAMLError as e:
        print(e)
        return None
    except FileNotFoundError:
        return None
    return payload


def round_trip_load_yaml(file_path):
    with open(file_path, 'r') as f:
        try:
            payload = yaml.round_trip_load(f)
        except yaml.YAMLError as e:
            print(e)
            return None
    return payload


def load_json(file_path):
    with open(file_path, 'r') as f:
        try:
            payload = json.load(f)
        except json.JSONDecodeError as e:
            print(e)
            return None
    return payload


def _init_jinja_env(env):
    def env_var(var, default=None):
        if var is None:
            return var
        return os.getenv(var, default)

    def as_bool(var):
        if var is None:
            return var
        return var.lower() in ('true', 'yes', '1')

    def as_number(var):
        if var is None:
            return var
        if var.isnumeric():
            return int(var)
        return float(var)

    def as_text(var):
        if var is None:
            return var
        return str(var)

    def silent_none(var):
        if var is None:
            return ""
        return var

    env.globals['env_var'] = env_var
    env.filters['as_bool'] = as_bool
    env.filters['as_number'] = as_number
    env.filters['as_text'] = as_text
    env.finalize = silent_none


def load_jinja_template(path: str):
    from jinja2 import Environment, FileSystemLoader

    env = Environment(loader=FileSystemLoader(searchpath=os.path.dirname(path)))
    _init_jinja_env(env)
    template = env.get_template(os.path.basename(path))

    return template


def load_jinja_string_template(value: str):
    from jinja2 import Environment, BaseLoader

    env = Environment(loader=BaseLoader())
    _init_jinja_env(env)
    template = env.from_string(value)

    return template


def open_report_in_browser(report_path='', is_cloud_path=False):
    protocol_prefix = "" if is_cloud_path else "file://"
    try:
        webbrowser.open(f"{protocol_prefix}{report_path}")
    except yaml.YAMLError as e:
        print(e)
        return None


def get_run_json_path(output_dir: str, input=None):
    console = Console()
    if input:
        if not os.path.exists(input):
            console.print(f'[bold red]Error: {input} not found[/bold red]')
            return
        if os.path.isdir(input):
            run_json = os.path.join(input, 'run.json')
        else:
            run_json = input
    else:
        latest = os.path.join(output_dir, 'latest')
        run_json = os.path.join(latest, 'run.json')
        if not os.path.isfile(run_json) and os.path.exists(output_dir):
            latest_report_dir_ctime = 0
            latest_report_dir = ''
            for de in os.scandir(output_dir):
                if not de.is_dir():
                    continue
                if os.path.getctime(de) > latest_report_dir_ctime:
                    latest_report_dir_ctime = os.path.getctime(de)
                    latest_report_dir = de.path
            run_json = os.path.join(output_dir, latest_report_dir, 'run.json')
    return os.path.abspath(run_json)



import math
import urllib.parse
from datetime import timedelta
from io import StringIO
from typing import Callable, Dict, List

from dbt.contracts.graph.manifest import Manifest

from piperider_cli.dbt import dbt_version
from piperider_cli.dbt.list_task import (
    list_changes_in_unique_id,
    list_modified_with_downstream, list_resources_data_from_manifest,
    load_manifest,
)
from piperider_cli.dbt.markdown import MarkdownTable
from piperider_cli.dbt.utils import ChangeType, ChangeUnit, ColumnChangeEntry, ColumnChangeView, JoinedTables, \
    MetricsChangeView, ResourceType


def embed_url_cli(content: str, url: str, unique_id: str, resource_type: str):
    return content


def embed_url_cloud(content: str, url: str, unique_id: str, resource_type: str):
    return f"[{content}]({url}?utm_source=pr#/{resource_type}s/{urllib.parse.quote(unique_id)})"


embed_url = None
try:
    from web import patch_sys_path_for_piperider_cli

    patch_sys_path_for_piperider_cli()
    embed_url = embed_url_cloud
except ImportError:
    embed_url = embed_url_cli


class DefaultChangeSetOpMixin:
    __slots__ = ()

    def load_run_as_manifest(self, run: Dict):
        manifest = run.get("dbt", {}).get("manifest", {})
        if manifest == {}:
            raise ValueError("Cannot find .dbt.manifest in run data")
        return load_manifest(manifest)

    def resolve_unique_id(self, resource_name: str, resource_type: str):
        for entry in self.base_resources + self.target_resources:
            if entry.get("resource_type") == resource_type and entry.get("name") == resource_name:
                return entry.get("unique_id")

    def has_changed(self, p1: Dict, p2: Dict):
        """
        p1 and p2 are table profiled data
        """

        base_cols: Dict[str, Dict] = p1.get("columns")
        target_cols: Dict[str, Dict] = p2.get("columns")

        if base_cols is None or target_cols is None:
            return True

        for k in base_cols:
            if ColumnChangeView(base_cols.get(k)) != ColumnChangeView(
                    target_cols.get(k)
            ):
                return True

        for k in target_cols:
            if ColumnChangeView(base_cols.get(k)) != ColumnChangeView(
                    target_cols.get(k)
            ):
                return True

        return False


class SummaryAggregate:
    display_name: str
    resource_type: str
    total: int

    explicit_changeset: List[ChangeUnit] = None

    # TODO discovery the downstream of added by "dbt list --select model_name+"
    # it mixed resources with "model" and "metric"
    modified_with_downstream: List[ChangeUnit] = None

    # value changes from profiling data
    diffs: List[str] = None
    no_diffs: List[str] = None
    skipped: List[str] = None

    def __init__(self, display_name: str):
        self.display_name = display_name
        if self.display_name == 'Models':
            self.resource_type = 'model'
        elif self.display_name == 'Metrics':
            self.resource_type = 'metric'
        elif self.display_name == 'Seeds':
            self.resource_type = 'seed'
        else:
            raise ValueError(f'Unknown type for {display_name}')

    @property
    def explicit_changes(self) -> int:
        """
        the code changes
        """
        self._ensure_initialized()
        return len(self.explicit_changeset)

    @property
    def impacted(self) -> int:
        """
        any code changes and all downstream
        """
        self._ensure_initialized()

        # TODO discovery the downstream of added by "dbt list --select model_name+"

        code_changes: List[str] = [x.unique_id for x in self.explicit_changeset]
        modified: List[str] = [x.unique_id for x in self.modified_with_downstream if
                               x.resource_type.value == self.resource_type]
        return len(set(modified + code_changes))

    @property
    def implicit_changes(self) -> int:
        """
        all value changes excluding the code changes
        """
        self._ensure_initialized()
        code_changes: List[str] = [x.unique_id for x in self.explicit_changeset]
        return len(set(self.diffs) - set(code_changes))

    @property
    def implicit_changeset(self) -> List[ChangeUnit]:
        self._ensure_initialized()
        code_changes: List[str] = [x.unique_id for x in self.explicit_changeset]
        ids = set(self.diffs) - set(code_changes)

        def _t(x: ChangeUnit):
            x.change_type = ChangeType.IMPLICIT
            return x

        return [_t(x) for x in self.modified_with_downstream if x.unique_id in ids]

    def _ensure_initialized(self):
        if self.explicit_changeset is None:
            raise ValueError('explicit_changeset is not initialized')
        if self.modified_with_downstream is None:
            raise ValueError('modified_with_downstream is not initialized')


class LookUpTable:

    def __init__(self, c: "SummaryChangeSet"):
        self.c: "SummaryChangeSet" = c

        self.path_mapping = self._build_path_mapping()
        self.base_execution, self.target_execution = self._build_dbt_time_mapping()
        self.base_tests, self.target_tests = self._build_tests_mapping()
        self.label_mapping = self._build_metric_label_mapping()
        self.grain_metric_mapping = self._build_grain_metric_mapping()
        self._sorted_weights: Dict[str, int] = self._build_sorting_parameters()

    def _build_path_mapping(self) -> Dict[str, str]:
        m = dict()
        for x in self.c.target_resources + self.c.base_resources:
            unique_id, resource_path = x.get('unique_id'), x.get('original_file_path')
            if unique_id not in m:
                m[unique_id] = resource_path

        return m

    def path(self, unique_id: str):
        p = self.path_mapping.get(unique_id)
        limit = 28
        if p is not None and len(p) >= limit:
            return f"..{p[len(p) - limit:]}"
        return p

    def tests(self, unique_id: str):
        return sum(self.base_tests.get(unique_id, dict(passed=0, failed=0)).values()), \
            sum(self.target_tests.get(unique_id, dict(passed=0, failed=0)).values())

    def failed_tests(self, unique_id: str):
        return self.base_tests.get(unique_id, dict(passed=0, failed=0)).get('failed'), \
            self.target_tests.get(unique_id, dict(passed=0, failed=0)).get('failed')

    def _build_dbt_time_mapping(self):
        base_execution = {x.get('unique_id'): x.get('execution_time') for x in
                          self.c.base.get('dbt', {}).get('run_results', {}).get('results', [])}
        target_execution = {x.get('unique_id'): x.get('execution_time') for x in
                            self.c.target.get('dbt', {}).get('run_results', {}).get('results', [])}

        return base_execution, target_execution

    def execution_time(self, unique_id: str):
        return self.base_execution.get(unique_id), self.target_execution.get(unique_id)

    @staticmethod
    def to_human_readable(seconds: float):
        if seconds is None:
            return "-"

        return str(timedelta(seconds=seconds))[:-4]

    def _build_tests_mapping(self):
        # TODO the test should be ref by the unique_id
        from collections import Counter
        # convert test results to (table_name, pass_or_not): count form
        b = Counter([(x.get('table'), x.get('status')) for x in self.c.base.get('tests', [])])
        t = Counter([(x.get('table'), x.get('status')) for x in self.c.target.get('tests', [])])

        def as_dict(c: Counter):
            m = dict()
            for (table, status), v in c.items():
                if table not in m:
                    m[table] = dict(passed=0, failed=0)

                if status == 'passed':
                    m[table]['passed'] += v
                else:
                    m[table]['failed'] += v
            return m

        return as_dict(b), as_dict(t)

    def _build_metric_label_mapping(self) -> Dict[str, str]:
        m = dict()
        for unique_id, v in self.c.base_manifest.metrics.items():
            m[unique_id] = v.label

        for unique_id, v in self.c.target_manifest.metrics.items():
            m[unique_id] = v.label

        return m

    def _build_grain_metric_mapping(self):
        mapping = dict()

        base_run_metrics = self.c.base.get('metrics', [])
        target_run_metrics = self.c.target.get('metrics', [])

        for unique_id in self.c.base_manifest.metrics:
            if unique_id not in mapping:
                mapping[unique_id] = dict()
            metrics_group = '.'.join(unique_id.split('.')[2:])
            for metric in base_run_metrics:
                if metrics_group == metric.get("headers")[1]:
                    name = metric.get("name")
                    if name not in mapping[unique_id]:
                        mapping[unique_id][name] = MetricsChangeView(name)
                    m = mapping[unique_id][name]
                    m.metric_group = metrics_group
                    m.base_data = metric.get("data")

        for unique_id in self.c.target_manifest.metrics:
            if unique_id not in mapping:
                mapping[unique_id] = dict()
            metrics_group = '.'.join(unique_id.split('.')[2:])
            for metric in target_run_metrics:
                if metrics_group == metric.get("headers")[1]:
                    name = metric.get("name")
                    if name not in mapping[unique_id]:
                        mapping[unique_id][name] = MetricsChangeView(name)
                    m = mapping[unique_id][name]
                    m.metric_group = metrics_group
                    m.target_data = metric.get("data")

        return mapping

    def _build_sorting_parameters(self):
        from piperider_cli import dbtutil
        from piperider_cli.dbt.sorting import topological_sort

        g1 = dbtutil.prepare_topological_graph(self.c.base.get('dbt', {}).get('manifest', {}))
        g2 = dbtutil.prepare_topological_graph(self.c.target.get('dbt', {}).get('manifest', {}))

        g = dict()
        for k in set(list(g1.keys()) + list(g2.keys())):
            if k in g1 and k in g2:
                g[k] = list(set(g1[k] + g2[k]))
                continue

            if k in g1:
                g[k] = g1[k]
            if k in g2:
                g[k] = g2[k]

        sorted_parameters = topological_sort(g, len(list(g.keys())))
        weights = dict()
        for x in sorted_parameters:
            weights[x] = len(weights)
        return weights

    def sort(self, changeset: List[ChangeUnit]) -> List[ChangeUnit]:
        from functools import cmp_to_key

        def callback(a: ChangeUnit, b: ChangeUnit) -> int:
            av = self._sorted_weights[a.unique_id]
            bv = self._sorted_weights[b.unique_id]

            if av == bv:
                return 0

            if av < bv:
                return -1

            if av > bv:
                return 1

        return sorted(changeset, key=cmp_to_key(callback))


class SummaryChangeSet(DefaultChangeSetOpMixin):
    def __init__(self, base: Dict, target: Dict):
        self.base: Dict = base
        self.target: Dict = target
        self.base_manifest: Manifest = self.load_run_as_manifest(base)
        self.target_manifest: Manifest = self.load_run_as_manifest(target)
        self.tables = JoinedTables(self.base, self.target)

        # resources in this format [{unique_id, name, resource_type}]
        self.base_resources = list_resources_data_from_manifest(self.base_manifest)
        self.target_resources = list_resources_data_from_manifest(self.target_manifest)

        self.mapper = LookUpTable(self)

        self.modified_models_and_metrics_with_downstream = \
            list_modified_with_downstream(self.base_manifest, self.target_manifest)

        self.models = SummaryAggregate('Models')
        self.metrics = SummaryAggregate('Metrics')
        self.seeds = SummaryAggregate('Seeds')

        # TODO the execute should be merged into each aggregate
        self.execute()

    def get_url(self):
        if hasattr(self, 'url'):
            return self.url
        return ""

    def execute(self):
        self.models.explicit_changeset = self.find_explicit_changes('model')
        self.update_models_value_changes()

        self.metrics.explicit_changeset = self.find_explicit_changes('metric')
        self.update_metrics_value_changes()

        self.seeds.explicit_changeset = self.find_explicit_changes('seed')
        self.update_seeds_value_changes()

        # configure the modified + downstream
        modified_with_downstream = [ChangeUnit(unique_id=x.get('unique_id'), change_type=ChangeType.IGNORED,
                                               resource_type=ResourceType.of(x.get('resource_type')))
                                    for x in self.modified_models_and_metrics_with_downstream]
        self.models.modified_with_downstream = modified_with_downstream
        self.metrics.modified_with_downstream = modified_with_downstream
        self.seeds.modified_with_downstream = modified_with_downstream
        self.update_modified_with_downstream(modified_with_downstream)

        # update total values
        self.models.total = len([x for x in self.target_resources if x.get('resource_type') == 'model'])
        self.metrics.total = len([x for x in self.target_resources if x.get('resource_type') == 'metric'])

    def find_explicit_changes(self, resource_type_filter: str):
        base_resources = [x.get("unique_id") for x in self.base_resources if
                          x.get('resource_type') == resource_type_filter]
        target_resources = [x.get("unique_id") for x in self.target_resources if
                            x.get('resource_type') == resource_type_filter]

        resource_in_both = list(set(base_resources).intersection(target_resources))
        # exclude added and removed by intersection with common resources
        modified = [
            x.get("unique_id")
            for x in list_changes_in_unique_id(
                self.base_manifest, self.target_manifest, True
            )
        ]

        def as_unit(unique_id: str, change_type: ChangeType):
            return ChangeUnit(unique_id=unique_id, change_type=change_type,
                              resource_type=ResourceType.of(resource_type_filter))

        modified = [as_unit(x, ChangeType.MODIFIED) for x in list(set(modified).intersection(resource_in_both))]
        added = [as_unit(x, ChangeType.ADDED) for x in list(set(target_resources) - set(base_resources))]
        removed = [as_unit(x, ChangeType.REMOVED) for x in list(set(base_resources) - set(target_resources))]
        return added + removed + modified

    def update_models_value_changes(self):
        # list implicit changes and exclude added and removed tables
        diffs = []
        no_diffs = []
        skipped = []

        def _resolve_id(ref_id: str, table_name: str):
            if ref_id:
                if not ref_id.startswith("model."):
                    return None
                return ref_id
            else:
                resolved_id = self.resolve_unique_id(table_name, "model")
                assert resolved_id is not None
                return resolved_id

        for table_name, ref_id, b, t in self.tables.table_data_iterator():
            r1, r2 = self.tables.row_counts(table_name)
            c1, c2 = self.tables.column_counts(table_name)
            both_profiled = not math.isnan(r1) and not math.isnan(r2)
            if both_profiled:
                if r1 == r2 and c1 == c2 and not self.has_changed(b, t):
                    resolved_id = _resolve_id(ref_id, table_name)
                    if resolved_id:
                        no_diffs.append(resolved_id)
                else:
                    resolved_id = _resolve_id(ref_id, table_name)
                    if resolved_id:
                        diffs.append(resolved_id)
            else:
                if b and t and self.has_changed(b, t):
                    resolved_id = _resolve_id(ref_id, table_name)
                    if resolved_id:
                        diffs.append(resolved_id)
                else:
                    resolved_id = _resolve_id(ref_id, table_name)
                    if resolved_id:
                        if not math.isnan(r2):
                            no_diffs.append(resolved_id)
                        else:
                            skipped.append(resolved_id)

        self.models.diffs = list(set(diffs))
        self.models.no_diffs = list(set(no_diffs))
        self.models.skipped = list(set(skipped))

    def update_seeds_value_changes(self):
        # list implicit changes and exclude added and removed tables
        diffs = []
        no_diffs = []
        skipped = []

        def _resolve_id(ref_id: str, table_name: str):
            if ref_id:
                if not ref_id.startswith("seed."):
                    return None
                return ref_id
            else:
                resolved_id = self.resolve_unique_id(table_name, "seed")
                assert resolved_id is not None
                return resolved_id

        for table_name, ref_id, b, t in self.tables.table_data_iterator():
            r1, r2 = self.tables.row_counts(table_name)
            c1, c2 = self.tables.column_counts(table_name)
            both_profiled = not math.isnan(r1) and not math.isnan(r2)
            if both_profiled:
                if r1 == r2 and c1 == c2 and not self.has_changed(b, t):
                    resolved_id = _resolve_id(ref_id, table_name)
                    if resolved_id:
                        no_diffs.append(resolved_id)
                else:
                    resolved_id = _resolve_id(ref_id, table_name)
                    if resolved_id:
                        diffs.append(resolved_id)
            else:
                if b and t and self.has_changed(b, t):
                    resolved_id = _resolve_id(ref_id, table_name)
                    if resolved_id:
                        diffs.append(resolved_id)
                else:
                    resolved_id = _resolve_id(ref_id, table_name)
                    if resolved_id:
                        if not math.isnan(r2):
                            no_diffs.append(resolved_id)
                        else:
                            skipped.append(resolved_id)

        self.seeds.diffs = list(set(diffs))
        self.seeds.no_diffs = list(set(no_diffs))
        self.seeds.skipped = list(set(skipped))

    def update_metrics_value_changes(self):
        # exclude added and removed
        metrics_b = {x.get("name"): x for x in self.base.get("metrics", {})}
        metrics_t = {x.get("name"): x for x in self.target.get("metrics", {})}

        diffs = []
        no_diffs = []
        # NOTE: The non-queried metrics are not included in the skipped list, because they are not in run.json
        # TODO: it would be nice to include them in the skipped list
        skipped = []

        def _resolve_id(ref_id: str, metric):
            if ref_id:
                if not ref_id.startswith("metric."):
                    return None
                return ref_id
            else:
                # resolve the unique id for the legacy report
                metric_name = metric.get(x).get("headers")[-1]
                ref_id = self.resolve_unique_id(metric_name, "metric")
                assert ref_id is not None
                return ref_id

        for x in set(list(metrics_b.keys()) + list(metrics_t.keys())):
            if x in metrics_b and x in metrics_t:
                if metrics_b.get(x) != metrics_t.get(x):
                    ref_id = metrics_t.get(x).get("ref_id")
                    resolved_id = _resolve_id(ref_id, metrics_t)
                    if resolved_id:
                        diffs.append(ref_id)
                else:
                    ref_id = metrics_t.get(x).get("ref_id")
                    resolved_id = _resolve_id(ref_id, metrics_t)
                    if resolved_id:
                        no_diffs.append(ref_id)
            else:
                if x in metrics_t:
                    ref_id = metrics_t.get(x).get("ref_id")
                    ref_id = _resolve_id(ref_id, metrics_t)
                    if ref_id:
                        no_diffs.append(ref_id)
                elif x in metrics_b:
                    ref_id = metrics_b.get(x).get("ref_id")
                    ref_id = _resolve_id(ref_id, metrics_b)
                    if ref_id:
                        skipped.append(ref_id)

        self.metrics.diffs = list(set(diffs))
        self.metrics.no_diffs = list(set(no_diffs) - set(diffs))
        self.metrics.skipped = list(set(skipped) - set(diffs) - set(no_diffs))

    def update_modified_with_downstream(self, units: List[ChangeUnit]):
        explicit_changes = self.models.explicit_changeset + self.metrics.explicit_changeset + self.seeds.explicit_changeset
        implicit_changes = self.models.implicit_changeset + self.metrics.implicit_changeset + self.seeds.implicit_changeset
        changes = explicit_changes + implicit_changes

        for unit in units:
            for c in changes:
                if unit.unique_id == c.unique_id:
                    unit.change_type = c.change_type
                    break

    def generate_markdown(self):
        # ref: https://gist.github.com/popcornylu/7a9f68c1ea80f09ba9c780d2026ce71e
        output = StringIO()

        def out_func(message: str):
            print(message, file=output)

        out_func("")

        self.generate_impact_summary_section(out_func)
        self.generate_resource_impact_section(out_func)

        return output.getvalue()

    def generate_impact_summary_section(self, out: Callable[[str], None]) -> None:
        out("# Impact Summary")
        if self.get_url():
            out(f"[PipeRider Report]({self.get_url()}?utm_source=pr)")
        out("### Code Changes")
        mt = MarkdownTable(headers=['Added', 'Removed', 'Modified'])
        explicit_changes = self.models.explicit_changeset + self.metrics.explicit_changeset + self.seeds.explicit_changeset
        added = [x for x in explicit_changes if x.change_type == ChangeType.ADDED]
        removed = [x for x in explicit_changes if x.change_type == ChangeType.REMOVED]
        modified = [x for x in explicit_changes if x.change_type == ChangeType.MODIFIED]
        mt.add_row([len(added), len(removed), len(modified)])
        out(mt.build())

        out("### Resource Impact")
        mt = MarkdownTable(headers=['Potentially Impacted', 'Assessed', 'Impacted'])
        potentially_impacted = [x.unique_id for x in (self.models.modified_with_downstream + removed)]
        impacted = [x for x in list(set(self.models.diffs + self.metrics.diffs + self.seeds.diffs)) if
                    x in potentially_impacted]
        assessed_no_impacted = [x for x in list(set(self.models.no_diffs + self.metrics.no_diffs + self.seeds.no_diffs))
                                if x in potentially_impacted]

        mt.add_row([len(potentially_impacted),
                    f"{len(impacted) + len(assessed_no_impacted)} assessed, "
                    f"{len(potentially_impacted) - len(impacted) - len(assessed_no_impacted)} skipped",
                    len(impacted)])
        out(mt.build())

    def generate_resource_impact_section(self, out: Callable[[str], None]) -> None:
        out("# Resource Impact")
        self.generate_models_section(out)
        self.generate_metrics_section(out)

    def generate_models_section(self, out: Callable[[str], None]) -> None:
        out("### Models")
        m = self.models
        modified_with_downstream = [x for x in m.modified_with_downstream if x.resource_type == ResourceType.MODEL]
        removed = [x for x in m.explicit_changeset if x.change_type == ChangeType.REMOVED]
        changeset = self.mapper.sort(modified_with_downstream + removed)
        if not changeset:
            out("No changes detected")
            return

        column_header = """
        Columns <br> <img src="https://raw.githubusercontent.com/InfuseAI/piperider/main/images/icons/icon-diff-delta-plus%402x.png" width="10px"> <img src="https://raw.githubusercontent.com/InfuseAI/piperider/main/images/icons/icon-diff-delta-minus%402x.png" width="10px"> <img src="https://raw.githubusercontent.com/InfuseAI/piperider/main/images/icons/icon-diff-delta-explicit%402x.png" width="10px">
        """.strip()

        mt = MarkdownTable(
            headers=['&nbsp;&nbsp;&nbsp;', 'Model', 'Impact', column_header, 'Rows', 'Dbt Time', 'Failed Tests',
                     'All Tests'])

        def impact(c: ChangeUnit):
            impacted = self.models.diffs
            skipped = self.models.skipped
            assessed_no_impacted = self.models.no_diffs
            if c.unique_id in impacted:
                return "Impacted"
            elif c.unique_id in skipped:
                return "Skipped"
            elif c.unique_id in assessed_no_impacted:
                return "Assessed not impacted"

        def cols(c: ChangeUnit):
            counts = self.tables.column_counts(c.unique_id)
            if c.change_type == ChangeType.ADDED:
                _, t = counts
                return t
            if c.change_type == ChangeType.REMOVED:
                b, _ = counts
                return f"~~{b}~~"
            if c.change_type == ChangeType.MODIFIED or c.change_type == ChangeType.IMPLICIT:
                b, t = counts
                """
                example:
                16 ($\color{green}{\text{ 1 }}$ / $\color{red}{\text{ 1 }}$ / $\color{orange}{\text{ 5 }}$)
                """

                changes = list(self.tables.columns_changed_iterator(c.unique_id))

                def col_state(c: ColumnChangeEntry):
                    if c.base_view.data is not None and c.target_view.data is not None:
                        return ChangeType.MODIFIED
                    if c.base_view.data is None:
                        return ChangeType.ADDED
                    if c.target_view.data is None:
                        return ChangeType.REMOVED
                    raise ValueError("Cannot be here")

                stat = [col_state(x) for x in changes]
                from collections import Counter
                counter = Counter(stat)
                param = dict(value=t,
                             added=counter.get(ChangeType.ADDED, 0),
                             removed=counter.get(ChangeType.REMOVED, 0),
                             modified=counter.get(ChangeType.MODIFIED, 0),
                             )

                output = r"%(value)s ($\color{green}{\text{ %(added)s }}$ /" \
                         r" $\color{red}{\text{ %(removed)s }}$ /" \
                         r" $\color{orange}{\text{ %(modified)s }}$)" % param
                return output
            if c.change_type == ChangeType.IGNORED:
                b, t = counts
                return t

            return '-'

        def rows(c: ChangeUnit):
            rows = self.tables.row_counts(c.unique_id)
            if c.change_type == ChangeType.ADDED:
                _, t = rows
                return t
            if c.change_type == ChangeType.REMOVED:
                b, _ = rows
                return f"~~{b}~~"
            if c.change_type == ChangeType.MODIFIED or c.change_type == ChangeType.IMPLICIT:
                b, t = rows
                """
                example:
                997 $\color{green}{\text{ (↑ 20) }}$
                97 $\color{red}{\text{ (↓ -2) }}$
                """
                text = r'%(value)s $\color{%(color)s}{\text{ (%(sign)s %(diff)s) }}$'
                return ChangeType.MODIFIED.display_changes(b, t, text)
            if c.change_type == ChangeType.IGNORED:
                b, t = rows
                return t
            return '-'

        def dbt_time(c: ChangeUnit):
            b, t = self.mapper.execution_time(c.unique_id)
            if c.change_type == ChangeType.ADDED:
                return LookUpTable.to_human_readable(t)

            if c.change_type == ChangeType.REMOVED:
                return ""
            if c.change_type == ChangeType.MODIFIED or c.change_type == ChangeType.IMPLICIT or c.change_type == ChangeType.IGNORED:
                """
                example:
                0:00:00.16 $\color{green}{\text{ (↓ 0.05) }}$
                """
                text = r'%(value)s $\color{%(color)s}{\text{ (%(sign)s %(diff).2f) }}$'
                return ChangeType.MODIFIED.display_changes(b, t, text, converter=LookUpTable.to_human_readable,
                                                           negative_change=True)

            return '-'

        def failed_tests(c: ChangeUnit):
            _, all_t = self.mapper.tests(c.table_name)
            b, t = self.mapper.failed_tests(c.table_name)
            if c.change_type == ChangeType.ADDED:
                if all_t == 0:
                    return "-"
                return t
            if c.change_type == ChangeType.REMOVED:
                return ""
            if c.change_type == ChangeType.MODIFIED or c.change_type == ChangeType.IMPLICIT or c.change_type == ChangeType.IGNORED:
                if all_t == 0:
                    return "-"
                if b == t:
                    return f"{t}"
                text = r'%(value)s ($\color{%(color)s}{\text{ (%(sign)s %(diff)s) }}$)'
                return ChangeType.MODIFIED.display_changes(b, t, text, negative_change=True)
            return "-"

        def all_tests(c: ChangeUnit):
            b, t = self.mapper.tests(c.table_name)
            if c.change_type == ChangeType.ADDED:
                if t == 0:
                    return "-"
                return t
            if c.change_type == ChangeType.REMOVED:
                return ""
            if c.change_type == ChangeType.MODIFIED or c.change_type == ChangeType.IMPLICIT or c.change_type == ChangeType.IGNORED:
                if t == 0:
                    return "-"
                if b == t:
                    return f"{t}"
                text = r'%(value)s ($\color{%(color)s}{\text{ (%(sign)s %(diff)s) }}$)'
                return ChangeType.MODIFIED.display_changes(b, t, text)
            return "-"

        for c in changeset[:50]:
            mt.add_row(
                [c.change_type.icon_image_tag,
                 embed_url(self.mapper.path(c.unique_id), self.get_url(), c.unique_id, c.resource_type.value),
                 impact(c), cols(c), rows(c), dbt_time(c), failed_tests(c),
                 all_tests(c)])

        if len(changeset) > 50:
            remainings = len(changeset) - 50
            mt.add_row(
                ['', f'{remainings} more potentially impacted models', '', '', '', '', ''])

        out(mt.build())

    def generate_metrics_section(self, out: Callable[[str], None]) -> None:
        out("### Metrics")
        m = self.metrics
        modified_with_downstream = [x for x in m.modified_with_downstream if x.resource_type == ResourceType.METRIC]
        removed = [x for x in m.explicit_changeset if x.change_type == ChangeType.REMOVED]
        changeset = self.mapper.sort(modified_with_downstream + removed)
        if not changeset:
            out("No changes detected")
            return

        grain_metrics = self.mapper.grain_metric_mapping
        labels = self.mapper.label_mapping

        for c in changeset:
            for gm in grain_metrics[c.unique_id].values():
                gm.update_status()

        metrics_summary = {}
        for c in changeset[:50]:
            label = labels[c.unique_id]
            metrics_summary[label] = {
                'total': 0, 'no-changes': 0, 'edited': 0, 'added': 0, 'removed': 0,
                'state_icon': c.change_type.icon_image_tag,
            }

            for gm in grain_metrics[c.unique_id].values():
                metrics_summary[label][gm.change_type] += 1
                metrics_summary[label]['total'] += 1

        if not metrics_summary:
            out("")

        def latex_orange(text):
            if text is None:
                return ""
            if isinstance(text, str) and '%' in text:
                text = text.replace('%', '\%')
            return r'$\color{orange}{\text{ %s }}$' % str(text)

        mt = MarkdownTable(
            headers=['&nbsp;&nbsp;&nbsp;', 'Metric', 'Impact', f"Queries <br> total ({latex_orange('change')})"])

        def impact(c: ChangeUnit):
            impacted = self.metrics.diffs
            skipped = self.metrics.skipped
            assessed_no_impacted = self.metrics.no_diffs
            if c.unique_id in impacted:
                return "Impacted"
            elif c.unique_id in skipped:
                return "Skipped"
            elif c.unique_id in assessed_no_impacted:
                return "Assessed not impacted"
            else:
                return "Skipped"

        for c in changeset[:50]:
            label = labels[c.unique_id]
            entry = metrics_summary[label]
            chagned = f"({latex_orange(str(entry['edited']))})" if entry['edited'] > 0 else ""
            mt.add_row([
                entry['state_icon'],
                embed_url(label, self.get_url(), c.unique_id, c.resource_type.value),
                impact(c),
                f"{entry['total'] if entry['total'] > 0 else '-'} {chagned}"
            ])

        if len(changeset) > 50:
            remainings = len(changeset) - 50
            mt.add_row(
                ['', f'{remainings} more potentially impacted metrics', ''])

        out(mt.build())

    def build_path_mapping(self) -> Dict[str, str]:
        m = dict()

        for x in self.target_resources + self.base_resources:
            unique_id, resource_path = x.get('unique_id'), x.get('original_file_path')
            if unique_id not in m:
                m[unique_id] = resource_path

        return m


class GraphDataChangeSet(DefaultChangeSetOpMixin):
    def __init__(self, base: Dict, target: Dict):
        self.base: Dict = base
        self.target: Dict = target
        self.base_manifest: Manifest = self.load_run_as_manifest(base)
        self.target_manifest: Manifest = self.load_run_as_manifest(target)

        # resources in this format [{unique_id, name, resource_type}]
        self.base_resources = list_resources_data_from_manifest(self.base_manifest)
        self.target_resources = list_resources_data_from_manifest(self.target_manifest)

        self.explicit_changes = sorted(self._do_list_explicit_changes())

    def _do_list_explicit_changes(self):
        base_resources = [x.get("unique_id") for x in self.base_resources]
        target_resources = [x.get("unique_id") for x in self.target_resources]

        resource_in_both = list(set(base_resources).intersection(target_resources))
        # exclude added and removed by intersection with common resources
        output = [
            x.get("unique_id")
            for x in list_changes_in_unique_id(
                self.base_manifest, self.target_manifest, True
            )
        ]
        output = list(set(output).intersection(resource_in_both))

        return [x for x in output if not x.startswith('test.')]

    def list_explicit_changes(self):
        return self.explicit_changes

    def _metrics_implicit_changes(self):
        if dbt_version >= '1.6':
            return []

        # exclude added and removed
        metrics_b = {x.get("name"): x for x in self.base.get("metrics", {})}
        metrics_t = {x.get("name"): x for x in self.target.get("metrics", {})}

        diffs = []
        for x in set(list(metrics_b.keys()) + list(metrics_t.keys())):
            if x in metrics_b and x in metrics_t:
                if metrics_b.get(x) != metrics_t.get(x):
                    ref_id = metrics_t.get(x).get("ref_id")
                    if ref_id:
                        if not ref_id.startswith("metric."):
                            continue
                        diffs.append(ref_id)
                    else:
                        # resolve the unique id for the legacy report
                        metric_name = metrics_t.get(x).get("headers")[-1]
                        ref_id = self.resolve_unique_id(metric_name, "metric")
                        assert ref_id is not None
                        diffs.append(ref_id)

        implicit = list(set(diffs))
        return implicit

    def _table_implicit_changes(self):
        # list implicit changes and exclude added and removed tables

        diffs = []

        tables = JoinedTables(self.base, self.target)
        for table_name, ref_id, b, t in tables.table_data_iterator():
            r1, r2 = tables.row_counts(table_name)
            c1, c2 = tables.column_counts(table_name)
            both_profiled = not math.isnan(r1) and not math.isnan(r2)
            if both_profiled:
                if r1 == r2 and c1 == c2 and not self.has_changed(b, t):
                    pass
                else:
                    if ref_id:
                        if not ref_id.startswith("model.") and not ref_id.startswith("seed."):
                            continue
                        diffs.append(ref_id)
                    else:
                        resolved_id = self.resolve_unique_id(table_name, "model")
                        if not resolved_id:
                            resolved_id = self.resolve_unique_id(table_name, "seed")
                        assert resolved_id is not None
                        diffs.append(resolved_id)

        return diffs

    def list_implicit_changes(self):
        table_implicit = self._table_implicit_changes()
        metric_implicit = self._metrics_implicit_changes()

        return sorted(table_implicit + metric_implicit)



from typing import Dict


def build_columns(column_details: Dict):
    result = dict()
    for name, v in column_details.items():
        result[name] = dict(name=name, description=v.get('description'),
                            type='other', schema_type='other')
    return result


def reverse_to_run(manifest_data: Dict):
    from piperider_cli.dbt.list_task import list_resources_data_from_manifest, load_manifest
    base_data = dict(id="tmp-run", tables={}, created_at="2023-08-14T03:46:51.629668Z",
                     datasource=dict(name='empty', type='duckdb'))
    dbt = dict(manifest=manifest_data, run_results={})
    base_data['dbt'] = dbt

    tables = dict()
    nodes = manifest_data.get('nodes')

    for x in list_resources_data_from_manifest(load_manifest(manifest_data)):
        resource_type = x.get('resource_type')
        unique_id = x.get('unique_id')

        if unique_id not in nodes:
            continue

        if resource_type not in ['model', 'seed', 'source']:
            continue
        table_name = x.get('name')
        column_details: Dict[Dict] = nodes.get(unique_id).get('columns')
        tables[table_name] = dict(name=table_name, col_count=len(column_details), columns=build_columns(column_details),
                                  ref_id=unique_id)

    base_data["tables"] = tables
    return base_data



import io
from typing import Union


class DbtUtil:

    @staticmethod
    def get_dbt_project_path(dbt_project_dir: str = None, no_auto_search: bool = False,
                             recursive: bool = True) -> str:
        import piperider_cli.dbtutil as u
        return u.get_dbt_project_path(dbt_project_dir=dbt_project_dir, no_auto_search=no_auto_search,
                                      recursive=recursive)

    @staticmethod
    def read_dbt_resources(source: Union[str, io.TextIOWrapper, list]):
        import piperider_cli.dbtutil as u
        return u.read_dbt_resources(source=source)

    @staticmethod
    def load_dbt_project(path: str):
        import piperider_cli.dbtutil as u
        return u.load_dbt_project(path)

    @staticmethod
    def load_dbt_profile(path: str):
        import piperider_cli.dbtutil as u
        return u.load_dbt_profile(path)

    @staticmethod
    def load_credential_from_dbt_profile(dbt_profile, profile_name, target_name):
        import piperider_cli.dbtutil as u
        return u.load_credential_from_dbt_profile(dbt_profile=dbt_profile, profile_name=profile_name,
                                                  target_name=target_name)



import hashlib
import json
import math
import os
import shlex
import shutil
import subprocess
import sys
import uuid
from datetime import datetime
from enum import Enum
from typing import List, Optional

from rich import box
from rich.color import Color
from rich.console import Console
from rich.pretty import Pretty
from rich.progress import BarColumn, Column, MofNCompleteColumn, Progress, TextColumn, TimeElapsedColumn
from rich.style import Style
from rich.table import Table
from sqlalchemy import inspect
from sqlalchemy.exc import NoSuchTableError

import piperider_cli.dbtutil as dbtutil
from piperider_cli import clone_directory, convert_to_tzlocal, datetime_to_str, event, \
    raise_exception_when_directory_not_writable
from piperider_cli.assertion_engine import AssertionEngine
from piperider_cli.assertion_engine.recommender import RECOMMENDED_ASSERTION_TAG
from piperider_cli.configuration import Configuration, FileSystem, ReportDirectory
from piperider_cli.datasource import DataSource
from piperider_cli.datasource.unsupported import UnsupportedDataSource
from piperider_cli.error import PipeRiderInvalidDataSourceError, PipeRiderError, PipeRiderConnectorUnsupportedError
from piperider_cli.exitcode import EC_ERR_TEST_FAILED, EC_WARN_NO_PROFILED_MODULES
from piperider_cli.metrics_engine import MetricEngine, MetricEventHandler
from piperider_cli.profiler import ProfileSubject, Profiler, ProfilerEventHandler
from piperider_cli.statistics import Statistics
from piperider_cli.utils import create_link, remove_link


class RunEventPayload:

    def __init__(self):
        # all fields
        self.tables = 0
        self.columns = []
        self.rows = []
        self.dbt_command = ''
        self.passed_assertions = 0
        self.failed_assertions = 0
        self.passed_dbt_testcases = 0
        self.failed_dbt_testcases = 0
        self.build_in_assertions = 0
        self.custom_assertions = 0
        self.recommended_assertions = 0

    def to_dict(self):
        return self.__dict__


class RichProfilerEventHandler(ProfilerEventHandler):

    def __init__(self, tables):
        table_width, total_width = self._get_width(tables)
        total_column = TextColumn("{task.fields[coft]}", table_column=Column(width=total_width))
        text_column = TextColumn("{task.description}", table_column=Column(width=table_width))
        bar_column = BarColumn(bar_width=80, pulse_style=Style.from_color(Color.from_rgb(244, 164, 96)))
        mofn_column = MofNCompleteColumn(table_column=Column(width=5, justify="right"))
        time_elapsed_column = TimeElapsedColumn()

        self.progress = Progress(total_column, text_column, bar_column, mofn_column, time_elapsed_column)
        self.progress_started = False
        self.tasks = {}
        self.table_total = 0
        self.table_completed = 0

    def _get_width(self, tables):
        names = ['METADATA'] + tables
        return max([len(x) for x in names]), len(str(len(names))) * 2 + 2

    def handle_run_start(self, run_result):
        self.progress.start()

    def handle_run_progress(self, run_result, total, completed):
        self.table_total = total

    def handle_run_end(self, run_result):
        self.progress.stop()

    def handle_metadata_start(self):
        self.progress.start()
        padding = ' ' * (len(str(self.table_total)) - len(str(self.table_completed)))
        coft = f'[{padding}{self.table_completed}/{self.table_total}]'
        task_id = self.progress.add_task('METADATA', total=None, coft=coft)
        self.tasks['__metadata__'] = task_id

    def handle_metadata_progress(self, total, completed):
        task_id = self.tasks['__metadata__']
        self.progress.update(task_id, completed=completed, total=total)

    def handle_metadata_end(self):
        self.progress.stop()
        task_id = self.tasks['__metadata__']
        self.progress.remove_task(task_id)

    def handle_manifest_start(self):
        self.progress.start()
        padding = ' ' * (len(str(self.table_total)) - len(str(self.table_completed)))
        coft = f'[{padding}{self.table_completed}/{self.table_total}]'
        task_id = self.progress.add_task('DBT Manifest', total=None, coft=coft)
        self.tasks['__manifest__'] = task_id
        pass

    def handle_manifest_progress(self, total, completed):
        task_id = self.tasks['__manifest__']
        self.progress.update(task_id, completed=completed, total=total)
        pass

    def handle_manifest_end(self):
        self.progress.stop()
        task_id = self.tasks['__manifest__']
        self.progress.remove_task(task_id)
        pass

    def handle_table_start(self, table_name):
        self.progress.start()
        self.table_completed += 1
        padding = ' ' * (len(str(self.table_total)) - len(str(self.table_completed)))
        coft = f'[{padding}{self.table_completed}/{self.table_total}]'
        task_id = self.progress.add_task(table_name, total=None, coft=coft)
        self.tasks[table_name] = task_id

    def handle_table_progress(self, table_name, table_result, total, completed):
        task_id = self.tasks[table_name]
        self.progress.update(task_id, total=total, completed=completed)

    def handle_table_end(self, table_name, table_result):
        self.progress.stop()
        task_id = self.tasks[table_name]
        self.progress.remove_task(task_id)

    def handle_column_start(self, table_name, column_name):
        pass

    def handle_column_end(self, table_name, column_name, column_result):
        pass


class RichMetricEventHandler(MetricEventHandler):

    def __init__(self, metrics):
        max_metric_width, counting_width = self._get_width(metrics)
        total_column = TextColumn("{task.fields[coft]}", table_column=Column(width=counting_width))
        text_column = TextColumn("{task.description}", table_column=Column(width=max_metric_width))
        bar_column = BarColumn(bar_width=80, pulse_style=Style.from_color(Color.from_rgb(244, 164, 96)))
        mofn_column = MofNCompleteColumn(table_column=Column(width=5, justify="right"))
        time_elapsed_column = TimeElapsedColumn()

        self.progress = Progress(total_column, text_column, bar_column, mofn_column, time_elapsed_column)
        self.progress_started = False
        self.tasks = {}
        self.metric_total = 0
        self.metric_completed = 0

    @staticmethod
    def _get_width(metrics):
        return max([len(x) for x in metrics]), len(str(len(metrics))) * 2 + 2

    def handle_run_start(self):
        pass

    def handle_run_progress(self, total: int, completed: int):
        self.metric_total = total

    def handle_run_end(self):
        self.progress.stop()

    def handle_metric_start(self, metric: str):
        self.metric_completed += 1
        padding = ' ' * (len(str(self.metric_total)) - len(str(self.metric_completed)))
        coft = f'[{padding}{self.metric_completed}/{self.metric_total}]'
        task_id = self.progress.add_task(metric, total=None, coft=coft)
        self.tasks[metric] = task_id
        self.progress.start()

    def handle_metric_progress(self, metric: str, total: int, completed: int):
        if completed == 0:
            task_id = self.tasks[metric]
            self.progress.update(task_id, total=total)

    def handle_metric_end(self, metric: str):
        self.progress.stop()
        task_id = self.tasks[metric]
        self.progress.remove_task(task_id)

    def handle_param_query_start(self, metric: str, param: str):
        pass

    def handle_param_query_end(self, metric: str):
        task_id = self.tasks[metric]
        self.progress.update(task_id, advance=1)


def _filter_subject(name: str, includes: List[str], excludes: List[str]) -> bool:
    name = name.upper()
    if includes is not None:
        includes = [include.upper() for include in includes]
        if name not in includes:
            return False

    if excludes is not None:
        excludes = [exclude.upper() for exclude in excludes]
        if name in excludes:
            return False

    return True


def _execute_assertions(console: Console, engine, ds_name: str, output, profiler_result, created_at):
    # TODO: Implement running test cases based on profiling result
    assertion_engine = AssertionEngine(engine)
    assertion_engine.load_assertions(profiler_result)

    results = exceptions = []
    # Execute assertions
    if len(assertion_engine.assertions):
        console.rule('Testing')
        results, exceptions = assertion_engine.evaluate_all()
    return results, exceptions


def _show_dbt_test_result(dbt_test_results, title=None, failed_only=False):
    console = Console()
    ascii_table = Table(show_header=True, show_edge=True, header_style='bold magenta',
                        box=box.SIMPLE, title=title)
    ascii_table.add_column('Status', style='bold white')
    ascii_table.add_column('Test Subject', style='bold')
    ascii_table.add_column('Assertion', style='bold green')
    ascii_table.add_column('Message', style='bold')

    for r in dbt_test_results:
        if failed_only and r.get('status') == 'passed':
            continue
        success = True if r.get('status') == 'passed' else False
        test_name = r.get('display_name')
        table = r.get('table')
        column = r.get('column')
        target = f'[yellow]{table}[/yellow]'
        if column:
            target = f'{target}.[blue]{column}[/blue]'
        message = r.get('message')

        if success:
            ascii_table.add_row(
                '[[bold green]  OK  [/bold green]]',
                target,
                test_name,
                message
            )
        else:
            ascii_table.add_row(
                '[[bold red]FAILED[/bold red]]',
                target,
                test_name,
                message
            )

    if ascii_table.rows:
        console.print(ascii_table)


def _show_assertion_result(results, exceptions, failed_only=False, single_table=None, title=None):
    console = Console()

    def _wrap_pretty(obj):
        if obj is None:
            return '-'
        return obj if isinstance(obj, str) else Pretty(obj)

    if results:
        ascii_table = Table(show_header=True, show_edge=True, header_style='bold magenta',
                            box=box.SIMPLE, title=title)
        ascii_table.add_column('Status', style='bold white')
        ascii_table.add_column('Test Subject', style='bold')
        ascii_table.add_column('Assertion', style='bold green')
        ascii_table.add_column('Expected', style='bold')
        ascii_table.add_column('Actual', style='cyan')

        for assertion in results:
            if single_table and single_table != assertion.table:
                continue
            if failed_only and assertion.result.status():
                continue
            table = assertion.table
            column = assertion.column
            test_function = assertion.result.name
            success = assertion.result.status()
            target = f'[yellow]{table}[/yellow].[blue]{column}[/blue]' if column else f'[yellow]{table}[/yellow]'
            if success:
                ascii_table.add_row(
                    '[[bold green]  OK  [/bold green]]',
                    target,
                    test_function,
                    _wrap_pretty(assertion.result.expected),
                    _wrap_pretty(assertion.result.actual)
                )
            else:
                ascii_table.add_row(
                    '[[bold red]FAILED[/bold red]]',
                    target,
                    test_function,
                    _wrap_pretty(assertion.result.expected),
                    _wrap_pretty(assertion.result.actual)
                )
                if assertion.result.exception:
                    msg = f'[grey11 on white][purple4]{type(assertion.result.exception).__name__}[/purple4](\'{assertion.result.exception}\')[/grey11 on white]'
                    ascii_table.add_row('', '', '', msg)

        if ascii_table.rows:
            console.print(ascii_table)
    # TODO: Handle exceptions
    pass


def _transform_assertion_result(table: str, results):
    tests = []
    columns = {}
    if results is None:
        return dict(tests=tests, columns=columns)

    for r in results:
        if r.table == table:
            entry = r.to_result_entry()
            if r.column:
                if r.column not in columns:
                    columns[r.column] = []
                columns[r.column].append(entry)
            else:
                tests.append(entry)

    return dict(tests=tests, columns=columns)


class PreRunValidatingResult(Enum):
    OK = 0
    ERROR = 1
    FAILED_TO_LOAD_CONNECTOR = 2
    FAILED_TO_CONNECT_DATASOURCE = 3


def _pre_run_validating(ds: DataSource) -> (PreRunValidatingResult, Exception):
    console = Console()
    err = ds.verify_connector()
    if err:
        return PreRunValidatingResult.FAILED_TO_LOAD_CONNECTOR, err

    try:
        ds.verify_connection()
    except Exception as err:
        return PreRunValidatingResult.FAILED_TO_CONNECT_DATASOURCE, err

    stop_runner = _validate_assertions(console)
    if stop_runner:
        return PreRunValidatingResult.ERROR, None

    return PreRunValidatingResult.OK, None


def _validate_assertions(console: Console):
    assertion_engine = AssertionEngine(None)
    assertion_engine.load_all_assertions_for_validation()
    results = assertion_engine.validate_assertions()
    # if results
    for result in results:
        # result
        console.print(f'  [[bold red]FAILED[/bold red]] {result.as_user_report()}')

    if results:
        # stop runner
        return True

    # continue to run profiling
    return False


def prepare_default_output_path(filesystem: ReportDirectory, created_at, ds):
    latest_symlink_path = os.path.join(filesystem.get_output_dir(), 'latest')
    latest_source = f"{ds.name}-{convert_to_tzlocal(created_at).strftime('%Y%m%d%H%M%S')}"
    output_path = os.path.join(filesystem.get_output_dir(), latest_source)

    if not os.path.exists(output_path):
        os.makedirs(output_path, exist_ok=True)

    # Create a symlink pointing to the latest output directory
    remove_link(latest_symlink_path)

    console = Console()
    if not os.path.exists(latest_symlink_path):
        create_link(output_path, latest_symlink_path)
    else:
        console.print(f'[bold yellow]Warning: {latest_symlink_path} already exists[/bold yellow]')

    return output_path


def _append_descriptions(profile_result):
    for table_v in profile_result['tables'].values():
        table_v['description'] = 'Description: N/A'
        for column_v in table_v['columns'].values():
            column_v['description'] = 'Description: N/A'


def _clean_up_profile_null_properties(table_results):
    if not table_results:
        return

    removed = []
    for t_metric, t_metric_val in table_results.items():
        if t_metric_val is None:
            removed.append(t_metric)
        elif isinstance(t_metric_val, float) and not math.isfinite(t_metric_val):
            table_results[t_metric] = str(t_metric_val)

    for r in removed:
        del table_results[r]

    removed = []
    for col_name, props in table_results.get('columns', {}).items():
        for k, v in props.items():
            if v is None:
                removed.append(dict(col=col_name, key=k))
            elif isinstance(v, float) and not math.isfinite(v):
                table_results['columns'][col_name][k] = str(v)

    for r in removed:
        del table_results['columns'][r['col']][r['key']]


def _append_descriptions_from_assertion(profile_result):
    engine = AssertionEngine(None)
    engine.load_assertion_content()
    for table_name, table_v in engine.assertions_content.items():
        if table_name not in profile_result['tables'] or table_v is None:
            continue
        table_desc = table_v.get('description', '')
        if table_desc:
            profile_result['tables'][table_name]['description'] = f'{table_desc}'

        columns_content = table_v.get('columns') if table_v.get('columns') else {}
        for column_name, column_v in columns_content.items():
            if column_name not in profile_result['tables'][table_name]['columns'] or column_v is None:
                continue
            column_desc = column_v.get('description', '')
            if column_desc:
                profile_result['tables'][table_name]['columns'][column_name][
                    'description'] = f'{column_desc}'


def _analyse_and_log_run_event(profiled_result, assertion_results, dbt_test_results):
    tables = profiled_result.get('tables', [])
    tables = {k: v for k, v in tables.items() if v}
    event_payload = RunEventPayload()
    event_payload.tables = len(tables)

    # Table info
    for t in tables:
        table = profiled_result['tables'][t]
        # null col_count when the metadata is not available
        event_payload.columns.append(table.get('col_count'))
        # null row_count when the table is not profiled
        event_payload.rows.append(table.get('row_count'))

    # Count PipeRider assertions
    for r in assertion_results or []:
        if r.is_builtin:
            event_payload.build_in_assertions += 1
        else:
            event_payload.custom_assertions += 1

        if RECOMMENDED_ASSERTION_TAG in r.tags:
            event_payload.recommended_assertions += 1

        if r.result.status():
            event_payload.passed_assertions += 1
        else:
            event_payload.failed_assertions += 1

    # Count dbt-test cases
    if dbt_test_results:
        for r in dbt_test_results:
            if r.get('status') == 'passed':
                event_payload.passed_dbt_testcases += 1
            else:
                event_payload.failed_dbt_testcases += 1

    event.log_event(event_payload.to_dict(), 'run')


def decorate_with_metadata(profile_result: dict):
    from piperider_cli import __version__
    from piperider_cli.profiler.version import schema_version

    configuration = Configuration.instance()
    project_id = configuration.get_telemetry_id()

    profile_result['version'] = __version__
    profile_result['project_id'] = f'{project_id}'
    profile_result['user_id'] = f'{event._collector._user_id}'
    profile_result['metadata_version'] = schema_version()


def _check_assertion_status(assertion_results, assertion_exceptions):
    if assertion_exceptions and len(assertion_exceptions) > 0:
        return False

    if assertion_results:
        for assertion in assertion_results:
            if not assertion.result.status():
                return False

    return True


def check_dbt_manifest_compatibility(ds: DataSource, dbt_state_dir: str):
    database = ds.get_database()
    schema = ds.get_schema()

    manifest = dbtutil._get_state_manifest(dbt_state_dir)
    models = [x for x in manifest.get('nodes').values() if x.get('resource_type') in ['model']]

    # all model’s schema should be the prefix of target’s schema
    def filter_schema(model):
        model_schema = model.get('schema', '')
        return model_schema == schema or model_schema.startswith(f'{schema}_')

    filtered_models = list(filter(filter_schema, models))

    if len(models) != len(filtered_models):
        return False

    # at least one model uses the target’s database
    return len([x for x in models if x.get('database') == database])


def get_dbt_all_subjects(dbt_state_dir, options, filter_fn):
    def as_subjects(nodes):
        result = []
        for node in nodes:
            if "source" == node.get('resource_type'):
                # NOTE: In manifest.json, the source definition will include the 'identifier' field.
                #       If users do not specify an additional 'identifier' in dbt source schema,
                #       its value will be the same as the "name" field.
                name = node.get('name')
                table = node.get('identifier')  # there is no alias in the source definition
                schema = node.get('schema')
                database = node.get('database')
                ref_id = node.get("unique_id")
            else:
                name = node.get('name')
                table = node.get('alias')
                schema = node.get('schema')
                database = node.get('database')
                ref_id = node.get("unique_id")
            result.append(ProfileSubject(table, schema, database, name, ref_id))
        return result

    subjects = as_subjects(dbtutil.get_dbt_state_candidate(dbt_state_dir, options, select_for_metadata=True))
    Statistics().reset()

    tagged_subjects = as_subjects(dbtutil.get_dbt_state_candidate(dbt_state_dir, options))

    total = len(tagged_subjects)
    if not options.get('dbt_resources') and options.get('tag') is None:
        tagged_subjects = list(filter(filter_fn, tagged_subjects))
        Statistics().add_field('filter', total - len(tagged_subjects))

    return tagged_subjects, subjects


def get_dbt_profile_subjects(dbt_state_dir, options, filter_fn):
    tagged_subjects, _ = get_dbt_all_subjects(dbt_state_dir, options, filter_fn)
    return tagged_subjects


def get_dbt_state_dir(target_path, dbt_config, ds, skip_datasource_connection=False):
    if target_path is None or os.path.exists(target_path) is False:
        project_dir = dbt_config.get('projectDir')
        dbt_project = dbtutil.load_dbt_project(project_dir)
        target_path = dbt_project.get('target-path') if dbt_project.get('target-path') else 'target'
        if os.path.isabs(target_path) is False:
            parent = project_dir if os.path.isabs(project_dir) else os.path.join(FileSystem.WORKING_DIRECTORY,
                                                                                 project_dir)
            target_path = os.path.join(parent, target_path)

    if not dbtutil.is_dbt_state_ready(target_path):
        return None, f"[bold red]Error:[/bold red] No available 'manifest.json' under '{target_path}'"

    if os.environ.get('PIPERIDER_SKIP_TARGET_CHECK', None) != '1' and skip_datasource_connection is False:
        if not check_dbt_manifest_compatibility(ds, target_path):
            return None, f"[bold red]Error:[/bold red] Target mismatched. Please run 'dbt compile -t {dbt_config.get('target')}' to generate the new manifest, or set the environment variable 'PIPERIDER_SKIP_TARGET_CHECK=1' to skip the check."

    return target_path, None


def get_git_branch():
    # NOTE: Provide git branch information directly for the archived dbt project without .git folder
    git_branch = os.environ.get('PIPERIDER_GIT_BRANCH', None)
    git_sha = os.environ.get('PIPERIDER_GIT_SHA', None)
    if git_branch is not None and git_sha is not None:
        return git_branch, git_sha

    def _exec(command_line: str):
        cmd = shlex.split(command_line)
        proc = None
        try:
            proc = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=FileSystem.WORKING_DIRECTORY,
            )
            outs, errs = proc.communicate()
        except BaseException:
            if proc:
                proc.kill()
                outs, errs = proc.communicate()
            else:
                return ''

        if outs is not None:
            outs = outs.decode().strip()
        return outs

    git_branch = _exec('git branch --show-current')
    try:
        line = _exec(f'git log {git_branch} --pretty=oneline -n 1')
        git_sha = line.split(' ')[0]
    except BaseException:
        git_sha = ''
    return git_branch, git_sha


class Runner():
    @staticmethod
    def exec(datasource=None, table=None, output=None, skip_report=False, dbt_target_path: str = None,
             dbt_resources: Optional[dict] = None, dbt_select: tuple = None, dbt_state: str = None,
             report_dir: str = None, skip_datasource_connection: bool = False):
        console = Console()

        raise_exception_when_directory_not_writable(output)

        configuration = Configuration.instance()
        filesystem = configuration.activate_report_directory(report_dir=report_dir)
        datasources = {}
        datasource_names = []
        for ds in configuration.dataSources:
            datasource_names.append(ds.name)
            datasources[ds.name] = ds

        if len(datasource_names) == 0:
            if skip_datasource_connection is False:
                console.print(
                    "[[bold red]Error[/bold red]] Data source not found. "
                    "Please check your dbt 'profiles.yml' configuration.")
                return 1
            else:
                configuration.dbt['target'] = 'dev'
                configuration.dbt['profile'] = 'unsupported-data-source'
                datasources = {'dev': UnsupportedDataSource('Unknown', dbt=configuration.dbt)}
                datasource_names = ['dev']

        if datasource:
            ds_name = datasource
        else:
            # if global dbt config exists, use dbt profile target
            # else use the first datasource
            ds_name = configuration.dbt.get('target') if configuration.dbt else datasource_names[0]

        if ds_name not in datasource_names:
            console.print(f"[[bold red]Error[/bold red]] Datasource '{ds_name}' doesn't exist")
            console.print(f"Available datasources: {', '.join(datasource_names)}")
            return 1

        ds = datasources[ds_name]

        passed, reasons = ds.validate()
        if not passed:
            console.print(f"[bold red]Error:[/bold red] The credential of '{ds.name}' is not configured.")
            for reason in reasons:
                console.print(f'    {reason}')
            console.print(
                "[bold yellow]Hint:[/bold yellow]\n  Please execute command 'piperider init' to move forward.")
            return 1

        if not datasource and configuration.dbt is None and len(datasource_names) > 1:
            console.print(
                f"[bold yellow]Warning: multiple datasources found ({', '.join(datasource_names)}), using '{ds_name}'[/bold yellow]\n")

        console.print(f'[bold dark_orange]DataSource:[/bold dark_orange] {ds.name}')
        console.rule('Validating')
        result, err = _pre_run_validating(ds)
        if result is PreRunValidatingResult.ERROR:
            console.print('\n\n[[bold red]ERROR[/bold red]] Stop profiling, please fix the syntax errors above.')
            return 1
        elif result is PreRunValidatingResult.FAILED_TO_LOAD_CONNECTOR:
            if skip_datasource_connection is False:
                if isinstance(err, PipeRiderConnectorUnsupportedError):
                    console.print(f'[[bold red]Error[/bold red]] {err}')
                    raise err
                else:
                    console.print(
                        f'[[bold red]Error[/bold red]] Failed to load the \'{ds.credential.get("type")}\' connector.')
                    if isinstance(err, PipeRiderError):
                        console.print(f'Hint: {err.hint}')
                return 1
        elif result is PreRunValidatingResult.FAILED_TO_CONNECT_DATASOURCE:
            if skip_datasource_connection is False:
                console.print(
                    '[[bold red]ERROR[/bold red]] Unable to connect the data source, please check the configuration')
                if ds.type_name == 'unsupported':
                    data_source_type = ds.credential.get('type')
                    raise PipeRiderInvalidDataSourceError(data_source_type, FileSystem.dbt_profiles_path)
                return 1

        dbt_config = ds.args.get('dbt')
        dbt_manifest = None
        dbt_run_results = None

        if dbt_config:
            if not dbtutil.is_ready(dbt_config):
                console.log(
                    '[[bold red]ERROR:[/bold red]] DBT configuration is not completed, please check the config.yml')
                return sys.exit(1)
            dbt_target_path, err_msg = get_dbt_state_dir(dbt_target_path, dbt_config, ds, skip_datasource_connection)
            if err_msg:
                console.print(err_msg)
                return sys.exit(1)
            dbt_manifest = dbtutil.get_dbt_manifest(dbt_target_path)
            dbt_run_results = dbtutil.get_dbt_run_results(dbt_target_path)
            if dbt_select:
                # If the dbt_resources were already provided by environment variable PIPERIDER_DBT_RESOURCES, skip the dbt select
                dbt_resources = dbt_resources if dbt_resources else dbtutil.load_dbt_resources(dbt_target_path,
                                                                                               select=dbt_select,
                                                                                               state=dbt_state)
        console.print('everything is OK.')
        run_id = uuid.uuid4().hex
        created_at = datetime.utcnow()
        if skip_datasource_connection:
            engine = None
        else:
            engine = ds.get_engine_by_database()

        subjects: List[ProfileSubject]
        dbt_metadata_subjects: List[ProfileSubject] = None
        dbt_test_results = None

        if dbt_config:
            if dbtutil.is_dbt_run_results_ready(dbt_target_path):
                dbt_test_results = dbtutil.get_dbt_state_tests_result(dbt_target_path, table_filter=table)

        if table:
            if len(table.split('.')) == 2:
                schema, table_name = table.split('.')
                subjects = [ProfileSubject(table_name, schema)]
            else:
                subjects = [ProfileSubject(table, ds.get_schema())]
            Statistics().add_field('total', len(subjects))
        else:
            def filter_fn(subject: ProfileSubject):
                return _filter_subject(subject.name, configuration.includes, configuration.excludes)

            if dbt_config:
                options = dict(
                    view_profile=configuration.include_views,
                    dbt_resources=dbt_resources,
                    tag=dbt_config.get('tag')
                )
                subjects, dbt_metadata_subjects = get_dbt_all_subjects(dbt_target_path, options, filter_fn)
            else:
                table_names = inspect(engine).get_table_names()
                if configuration.include_views:
                    table_names += inspect(engine).get_view_names()

                subjects = [ProfileSubject(table_name) for table_name in table_names]
                subjects = list(filter(filter_fn, subjects))

        run_result = {}
        profiler_result = {}

        statistics = Statistics()
        # Profile the datasource
        profiler = Profiler(ds, RichProfilerEventHandler([subject.name for subject in subjects]), configuration)

        if skip_datasource_connection:
            # Generate run result from dbt manifest
            console.rule('Analyze dbt manifest')
            profiler_result = profiler.collect_metadata_from_dbt_manifest(dbt_manifest, dbt_metadata_subjects, subjects)
            console.rule('Skip Profile Data Source', style='dark_orange')
            run_result.update(profiler_result)
        else:
            try:
                console.rule('Collect metadata')
                profiler.collect_metadata(dbt_metadata_subjects, subjects)

                console.rule('Profile statistics')
                profiler_result = profiler.profile(subjects, metadata_subjects=dbt_metadata_subjects)
                run_result.update(profiler_result)
            except NoSuchTableError as e:
                console.print(f"[bold red]Error:[/bold red] No such table '{str(e)}'")
                return 1
            except Exception as e:
                raise Exception(f'Profiler Exception: {type(e).__name__}(\'{e}\')')

        statistics.reset()

        if skip_datasource_connection is False:
            console.rule('Query metrics')
            metrics = []
            if dbt_config:
                metrics = dbtutil.get_dbt_state_metrics_16(dbt_target_path, dbt_config.get('tag'), dbt_resources)
            statistics.display_statistic('query', 'metric')
            if metrics:
                run_result['metrics'] = MetricEngine(
                    ds,
                    metrics,
                    RichMetricEventHandler([m.label for m in metrics])
                ).execute()

        # TODO: refactor input unused arguments
        if skip_datasource_connection:
            assertion_results, assertion_exceptions = [], []
        else:
            assertion_results, assertion_exceptions = _execute_assertions(console, engine, ds.name, output,
                                                                          profiler_result, created_at)

        run_result['tests'] = []
        if assertion_results or dbt_test_results:
            console.rule('Assertion Results')
            if dbt_test_results:
                console.rule('dbt')
                _show_dbt_test_result(dbt_test_results)
                run_result['tests'].extend(dbt_test_results)
                if assertion_results:
                    console.rule('PipeRider')
            if assertion_results:
                _show_assertion_result(assertion_results, assertion_exceptions)
                run_result['tests'].extend([r.to_result_entry() for r in assertion_results])

        if not table:
            if dbt_config:
                run_result['dbt'] = dict()
                if dbt_manifest:
                    def _slim_dbt_manifest(manifest):
                        for key in manifest['nodes'].keys():
                            raw_code = manifest['nodes'][key]['raw_code']
                            sha1 = hashlib.sha1()
                            sha1.update(raw_code.encode('utf-8'))
                            manifest['nodes'][key]['raw_code'] = sha1.hexdigest()
                        return manifest

                    size = sys.getsizeof(dbt_manifest)
                    if size > 1024 * 1024 * 10:
                        # Reduce the manifest size if it's larger than 10MB
                        run_result['dbt']['manifest'] = _slim_dbt_manifest(dbt_manifest)
                    else:
                        run_result['dbt']['manifest'] = dbt_manifest
                if dbt_run_results:
                    run_result['dbt']['run_results'] = dbt_run_results

        for t in run_result['tables']:
            _clean_up_profile_null_properties(run_result['tables'][t])

        if dbt_config:
            dbtutil.append_descriptions(run_result, dbt_target_path)
        _append_descriptions_from_assertion(run_result)

        run_result['id'] = run_id
        run_result['created_at'] = datetime_to_str(created_at)
        git_branch, git_sha = get_git_branch()
        run_result['datasource'] = dict(name=ds.name, type=ds.type_name, git_branch=git_branch, git_sha=git_sha)

        decorate_with_metadata(run_result)

        output_path = prepare_default_output_path(filesystem, created_at, ds)
        output_file = os.path.join(output_path, 'run.json')

        with open(output_file, 'w', encoding='utf-8') as f:
            f.write(json.dumps(run_result, separators=(',', ':')))

        if dbt_config:
            abs_dir = os.path.abspath(dbt_target_path)
            dbt_state_files = ['manifest.json', 'run_results.json', 'index.html', 'catalog.json']
            dbt_output_dir = os.path.join(output_path, 'dbt')
            os.makedirs(dbt_output_dir, exist_ok=True)
            for file in dbt_state_files:
                abs_file_path = os.path.join(abs_dir, file)
                if not os.path.exists(abs_file_path):
                    continue
                shutil.copy2(abs_file_path, dbt_output_dir)

        if output:
            clone_directory(output_path, output)

        if skip_report:
            console.print(f'Results saved to {output if output else output_path}')

        _analyse_and_log_run_event(run_result, assertion_results, dbt_test_results)

        if not _check_assertion_status(assertion_results, assertion_exceptions):
            return EC_ERR_TEST_FAILED

        if len(subjects) == 0 and len(run_result.get('metrics', [])) == 0:
            return EC_WARN_NO_PROFILED_MODULES

        return 0



import json
import os
import sys
import webbrowser
from typing import List, Optional

import inquirer
import readchar
from rich import box
from rich.console import Console
from rich.prompt import Prompt
from rich.table import Table

from piperider_cli import datetime_to_str, open_report_in_browser, str_to_datetime, get_run_json_path
from piperider_cli.cloud import PipeRiderCloud, PipeRiderProject
from piperider_cli.compare_report import CompareReport, RunOutput
from piperider_cli.configuration import Configuration
from piperider_cli.datasource import FANCY_USER_INPUT
from piperider_cli.error import CloudReportError
from piperider_cli.githubutil import fetch_pr_metadata

console = Console()
piperider_cloud = PipeRiderCloud()

WORKSPACE_PROJECT_NAME_REGEX = r'^[a-zA-Z0-9-_]+$'


def _ask_email():
    if FANCY_USER_INPUT:
        account = inquirer.text('Email address', validate=lambda _, x: '@' in x)
    else:
        while True:
            account = Prompt.ask('[[yellow]?[/yellow]] Email address')
            if '@' in account:
                break
            else:
                console.print(f"'{account}' is not a valid")

    return account


def _ask_api_token():
    console.print('We have just sent an email with a magic link.\n'
                  'Please paste the api token from the magic link.')
    while True:
        api_token = Prompt.ask('[[yellow]?[/yellow]] API token')
        if len(api_token) > 0:
            break
    return api_token


def _process_magic_signup(account: str, username: str = None):
    response = piperider_cloud.magic_signup(account, username)
    if response is None:
        console.print('[[red]Error[/red]] Signup failed. Please try again.')
        return False
    elif response.get('success') is False:
        # TODO: use error code
        if 'User email or username already exists.' in response.get('message', ''):
            console.print(f"'{account}' already exists.")
            if not _process_magic_login(account):
                return False
        else:
            console.print(f"[[red]Error[/red]] {response.get('message')}")
            return False
    return True


def _process_magic_login(account: str):
    response = piperider_cloud.magic_login(account)
    if response is None:
        console.print('[[red]Error[/red]] Login failed. Please try again.')
        return False
    elif response.get('success') is False:
        # TODO: use error code
        if f"'{account}' does not exist." in response.get('message', ''):
            console.print('This email account was not found.')
            question = f"Signup with '{account}' (y/n)"
            allowed_ans = ['y', 'n', 'Y', 'N']
            if FANCY_USER_INPUT:
                ans = inquirer.text(question, validate=lambda _, x: x in allowed_ans)
            else:
                while True:
                    ans = Prompt.ask(f'[[yellow]?[/yellow]] {question}')
                    if ans in allowed_ans:
                        break
                    else:
                        console.print(f"'{ans}' is not a valid")
            if ans.upper() == 'N':
                return False
            else:
                if _process_magic_signup(account):
                    return True
                return False
        else:
            console.print(f"[[red]Error[/red]] {response.get('message')}")
            return False

    if response.get('link'):
        webbrowser.open(response.get('link'))
    return True


def ask_signup_info() -> str:
    console.print('Please provide your email account to signup')
    account = _ask_email()

    if _process_magic_signup(account):
        return _ask_api_token()
    return ''


def ask_login_info() -> str:
    console.print('Please provide your email account to login')
    account = _ask_email()

    if _process_magic_login(account):
        return _ask_api_token()
    return ''


def verify_api_token(api_token) -> bool:
    if piperider_cloud.validate(api_token) is False:
        console.print('[[red]Error[/red]] Invalid API Token. Please try again.')
        return False

    # Write API Token back to user profile
    piperider_cloud.service.update_api_token()
    return True


def location_mapper(loc_id):
    map = {
        'us': 'North America',
        'eu': 'Europe',
        'ap': 'Asia Pacific'
    }
    return map.get(loc_id, 'N/A')


def show_user_info():
    user = piperider_cloud.me
    if user is None:
        return
    ascii_table = Table(show_header=True, show_edge=True, header_style="bold magenta",
                        box=box.SIMPLE, title='User Profile')
    ascii_table.add_column('Email', justify='left', style='cyan')
    ascii_table.add_column('Username', justify='left')
    ascii_table.add_column('Full Name', justify='left')
    ascii_table.add_column('Storage Location', justify='left')
    ascii_table.add_column('Timezone', justify='left')
    ascii_table.add_row(
        user.get('email'),
        user.get('username', 'N/A'),
        user.get('fullname', 'N/A'),
        location_mapper(user.get('data_storage_location')),
        user.get('timezone_id')
    )
    console.print(ascii_table)
    pass


def setup_cloud_default_config(options: dict = None):
    if piperider_cloud.config.get('auto_upload') is None:
        piperider_cloud.update_config({'auto_upload': False})

    if options and options.get('default_project'):
        project_name = options.get('default_project')
        if piperider_cloud.get_project_by_name(project_name):
            piperider_cloud.update_config({'default_project': project_name})
            console.print(f'[[bold green]Config[/bold green]] Default project is set to \'{project_name}\'')
        else:
            console.print(f"[[yellow]Skip[/yellow]] Project '{project_name}' does not exist.")

    if piperider_cloud.config.get('default_project') is None:
        no_interaction = options and options.get('no_interaction')
        CloudConnector.select_project(no_interaction=no_interaction)


def select_reports(report_dir=None, datasource=None) -> List[RunOutput]:
    filesystem = Configuration.instance().activate_report_directory(report_dir=report_dir)
    selector = CompareReport(filesystem.get_output_dir(), None, None, datasource=datasource)
    console.rule('Select Reports to Upload')
    return selector.select_multiple_reports(action='upload')


class CloudReportOutput(RunOutput):
    def __init__(self, report):
        self.report = report
        self.id = report['id']
        self.name = report['datasource_name']
        self.created_at = report['created_at']
        self.table_count = report['tables']
        self.pass_count = report['passed']
        self.fail_count = report['failed']


def select_cloud_report_ids(datasource=None, project: PipeRiderProject = None, target=None,
                            base=None) -> (int, int):
    if project is None:
        project = piperider_cloud.get_default_project()

    if target:
        target_id = get_run_report_id(project, target)
    if base:
        base_id = get_run_report_id(project, base)

    reports = [CloudReportOutput(r) for r in piperider_cloud.list_reports(project, datasource=datasource)]
    if len(reports) == 0:
        return None, None

    selector = CompareReport(None, None, None, datasource=datasource, profiler_outputs=reports)

    if base is None and target is None:
        console.rule('Please select base and target reports to compare')
        base, target = selector.select_two_reports(action='compare')
        target_id = target.id
        base_id = base.id
    elif base and target is None:
        console.rule('Please select a target report to compare')
        target = selector.select_one_report(action='compare')
        target_id = target.id
    elif target and base is None:
        console.rule('Please select a base report to compare')
        base = selector.select_one_report(action='compare')
        base_id = base.id

    return base_id, target_id


def upload_to_cloud(run: RunOutput, debug=False, project: PipeRiderProject = None, show_progress=True) -> dict:
    response = piperider_cloud.upload_run(run.path, project=project, show_progress=show_progress)

    # TODO refine the output when API is ready

    def _patch_cloud_upload_response(run_path, project: PipeRiderProject, run_id):
        with open(run_path, 'r') as f:
            report = json.load(f)
        report['cloud'] = {
            'run_id': run_id,
            'project_name': f'{project.workspace_name}/{project.name}'
        }
        with open(run_path, 'w', encoding='utf-8') as f:
            f.write(json.dumps(report, separators=(',', ':')))

    if response.get('success') is True:
        run_id = response.get('id')
        report_url = response.get('url')
        if run_id:
            _patch_cloud_upload_response(run.path, project, run_id)

        return {
            'success': True,
            'message': response.get("message"),
            'run_id': run_id,
            'report_url': report_url,
            'file_path': run.path
        }

    if debug:
        console.print(response)
    return {
        'success': False,
        'message': response.get("message"),
        'report_url': 'N/A',
        'file_path': run.path
    }


def get_run_report_id(project: PipeRiderProject, report_key: str) -> Optional[int]:
    if report_key.isdecimal():
        if int(report_key) < 1:
            return None
        return int(report_key)

    if report_key.startswith('datasource:'):
        datasource = report_key.split(':')[-1]
        reports = piperider_cloud.list_reports(project, datasource=datasource)

        if reports:
            return reports[0].get('id')

    return None


def create_compare_reports(base_id: str, target_id: str, tables_from, project: PipeRiderProject = None,
                           metadata: dict = None) -> dict:
    if project is None:
        project = piperider_cloud.get_default_project()
    response = piperider_cloud.compare_reports(base_id, target_id, tables_from, project=project, metadata=metadata)
    return response


class CloudConnector:
    @staticmethod
    def is_login() -> bool:
        return piperider_cloud.available

    @staticmethod
    def is_auto_upload() -> bool:
        return piperider_cloud.config.get('auto_upload', False)

    @staticmethod
    def config_auto_upload(flag: bool):
        console.print(f'[[bold green]Config[/bold green]] Default auto upload behavior is set to {flag}')
        piperider_cloud.update_config({'auto_upload': flag})

    @staticmethod
    def signup(debug=False):

        if piperider_cloud.available:
            console.rule('Already Logged In')
            show_user_info()
            setup_cloud_default_config()
            return 0

        api_token = ask_signup_info()

        if api_token:
            if verify_api_token(api_token):
                console.rule('Login Successful')
                show_user_info()
                setup_cloud_default_config()
                return 0

        console.rule('Login Failed', style='red')
        return 1

    @staticmethod
    def login(api_token=None, options: dict = None, debug=False):

        if piperider_cloud.available:
            console.rule('Already Logged In')
            show_user_info()
            setup_cloud_default_config(options)
            return 0

        if api_token is None:
            api_token = ask_login_info()

        if api_token:
            if verify_api_token(api_token):
                console.rule('Login Successful')
                show_user_info()
                setup_cloud_default_config(options)
                return 0

        console.rule('Login Failed', style='red')
        return 1

    @staticmethod
    def logout():

        if piperider_cloud.available is False:
            console.rule('Already Logout', style='yellow')
            return 0

        piperider_cloud.logout()
        console.rule('Logout')
        return 0

    @staticmethod
    def upload_latest_report(report_dir=None, debug=False, open_report=False, enable_share=False,
                             project_name: str = None) -> int:
        filesystem = Configuration.instance().activate_report_directory(report_dir=report_dir)
        latest_report_path = get_run_json_path(filesystem.get_output_dir())
        return CloudConnector.upload_report(latest_report_path, debug=debug, open_report=open_report,
                                            enable_share=enable_share, project_name=project_name)

    @staticmethod
    def upload_report(report_path=None, report_dir=None, datasource=None, debug=False, open_report=False,
                      enable_share=False, project_name=None, show_progress=True) -> int:
        if piperider_cloud.available is False:
            console.rule('Please login PipeRider Cloud first', style='red')
            return 1

        if project_name:
            project = piperider_cloud.get_project_by_name(project_name)
            if project is None:
                console.print(f'[[bold red]Error[/bold red]] Project \'{project_name}\' does not exist')
                return 1
        else:
            project = piperider_cloud.get_default_project()

        rc = 0
        results = []
        if report_path is None:
            reports = select_reports(report_dir=report_dir, datasource=datasource)
            if show_progress:
                console.rule('Uploading Reports')
            for r in reports:
                response = upload_to_cloud(r, debug, project=project, show_progress=show_progress)
                if response.get('success') is False:
                    rc = 1
                response['name'] = r.name
                response['created_at'] = r.created_at
                results.append(response)
        else:
            if show_progress:
                console.rule('Uploading Report')
            report = RunOutput(report_path)
            response = upload_to_cloud(report, debug, project=project, show_progress=show_progress)
            if response.get('success') is False:
                rc = 1
            response['name'] = report.name
            response['created_at'] = report.created_at
            results.append(response)

        if show_progress:
            console.rule('Upload Completed')
            ascii_table = Table(show_header=True, show_edge=True, header_style="bold magenta",
                                box=box.SIMPLE)
            ascii_table.add_column('ID', justify='left', style='cyan')
            ascii_table.add_column('Status', justify='left', style='cyan')
            ascii_table.add_column('Name', justify='left')
            ascii_table.add_column('Created At', justify='left')
            ascii_table.add_column('Message', justify='left')

            reports = []
            for response in results:
                status = '[bold green]Success[/bold green]' if response.get(
                    'success') else '[bold yellow]Skipped[/bold yellow]'
                reports.append((response.get('run_id'), response.get('report_url', 'N/A')))
                message = response.get('message')
                created_at = datetime_to_str(str_to_datetime(response.get('created_at')), to_tzlocal=True)
                ascii_table.add_row(str(response.get('run_id')), status, response.get('name'), created_at, message)
            console.print(ascii_table)

            for report in reports:
                console.print(f'Report #{report[0]} URL: [deep_sky_blue1]{report[1]}?utm_source=cli[/deep_sky_blue1]', soft_wrap=True)

        if open_report:
            url = response.get('report_url')
            open_report_in_browser(url, True)

        if enable_share:
            run_id = response.get('run_id')
            CloudConnector.share_run_report(run_id, debug, project_name=project_name)

        return rc

    @staticmethod
    def generate_compare_report(base_id: str, target_id: str, tables_from='all',
                                project_name: str = None, debug: bool = False):
        # TODO: Change to use new front-end URL pattern
        def _generate_legacy_compare_report_url(base_id, target_id, project=None):
            metadata = None
            if project is None:
                project = piperider_cloud.get_default_project()
            if os.environ.get('GITHUB_EVENT_PATH'):
                metadata = fetch_pr_metadata()

            response = create_compare_reports(base_id, target_id, tables_from, project=project, metadata=metadata)
            return response

        try:
            # TODO: change to use workspace and project name instead of project id
            if project_name:
                project = piperider_cloud.get_project_by_name(project_name)
                if project is None:
                    console.print(f'[[bold red]Error[/bold red]] Project \'{project_name}\' does not exist')
                    return 1
            else:
                project = piperider_cloud.get_default_project()
            return _generate_legacy_compare_report_url(base_id, target_id, project=project)
        except CloudReportError as e:
            console.print(f'[[bold red]Error[/bold red]] Cannot generate compare report in the cloud: {e.message}')
            return None

    @staticmethod
    def compare_reports(base=None, target=None, tables_from='all', summary_file=None, debug=False,
                        project_name=None) -> int:
        if piperider_cloud.available is False:
            console.rule('Please login PipeRider Cloud first', style='red')
            return 1

        if project_name:
            project = piperider_cloud.get_project_by_name(project_name)
            if project is None:
                console.print(f'[[bold red]Error[/bold red]] Project \'{project_name}\' does not exist')
                return 1
        else:
            project = piperider_cloud.get_default_project()

        base_id, target_id = select_cloud_report_ids(base=base, target=target, project=project)
        if base_id is None or target_id is None:
            raise Exception(
                f'No reports found in the project: {project.workspace_name}/{project.name}. '
                f'Please upload reports to PipeRider Cloud first.')

        console.print(f"Creating comparison report id={base_id} ... id={target_id}")

        metadata = None
        if os.environ.get('GITHUB_EVENT_PATH'):
            metadata = fetch_pr_metadata()

        response = create_compare_reports(base_id, target_id, tables_from, project=project, metadata=metadata)
        if response is None:
            console.print('Failed to create the comparison report')
        else:
            url = response.get('url')
            console.print(f'Comparison report URL: {url}?utm_source=cli')

        if debug:
            console.print(response)

        if summary_file:
            summary_file = os.path.abspath(summary_file)
            summary_dir = os.path.dirname(summary_file)
            if summary_dir:
                os.makedirs(summary_dir, exist_ok=True)
            with open(summary_file, 'w', encoding='utf-8') as f:
                f.write(response.get('summary'))

    @staticmethod
    def share_run_report(run_id=None, debug=False, project_name=None):
        if piperider_cloud.available is False:
            console.rule('Please login PipeRider Cloud first', style='red')
            return 1

        if project_name is None:
            workspace_name, project_name = piperider_cloud.get_default_workspace_and_project()
        else:
            workspace_name, project_name = project_name.split('/')

        if workspace_name is None or project_name is None:
            console.rule('Please select a workspace and a project first', style='red')
            return 1

        if debug:
            console.print(f"Sharing comparison report id={run_id}")
        piperider_cloud.share_run_report(
            workspace_name=workspace_name,
            project_name=project_name,
            run_id=run_id
        )

    @staticmethod
    def share_compare_report(base_id=None, target_id=None, debug=False, project_name=None):
        if piperider_cloud.available is False:
            console.rule('Please login PipeRider Cloud first', style='red')
            return 1

        if project_name is None:
            workspace_name, project_name = piperider_cloud.get_default_workspace_and_project()
        else:
            workspace_name, project_name = project_name.split('/')

        if workspace_name is None or project_name is None:
            console.rule('Please select a workspace and a project first', style='red')
            return 1

        if debug:
            console.print(f"Sharing comparison report id={base_id} ... id={target_id}")
        piperider_cloud.share_compare_report(
            workspace_name=workspace_name,
            project_name=project_name,
            base_id=base_id,
            target_id=target_id,
        )

    @staticmethod
    def list_projects(debug=False) -> int:
        if piperider_cloud.available is False:
            console.rule('Please login PipeRider Cloud first', style='red')
            return 1

        projects = piperider_cloud.list_projects()
        # console.print(projects)

        layout_table = Table(
            title="PipeRider Cloud - Project List",
            title_style='bold magenta',
            show_header=True,
            show_edge=True,
            box=box.SIMPLE_HEAVY,
        )
        layout_table.add_column('Name')
        layout_table.add_column('Workspace')
        # layout_table.add_column('Project URL', justify='right', no_wrap=True)

        for project in projects:
            # TODO: Put the project URL in the table once the project URL is available
            # project_id = project.get('id')
            # project_url = f'[deep_sky_blue1]{piperider_cloud.service.cloud_host}/projects/{project_id}[/deep_sky_blue1]'
            project_name = project.get('name') if project.get(
                'workspace_name') is None else f"{project.get('workspace_name')}/{project.get('name')}"
            workspace_name = project.get('workspace_name')
            display_name = project.get('workspace_display_name') if project.get(
                'workspace_display_name') else workspace_name
            layout_table.add_row(
                project_name,
                display_name
                # project_url,
            )

        console.print(layout_table)
        pass

    @staticmethod
    def select_project(project_name: str = None,
                       datasource: str = None,
                       no_interaction: bool = False,
                       debug: bool = False) -> int:

        def _select_project_by_inquirer(project_list: list) -> PipeRiderProject:
            arrow_alias_msg = ''
            if sys.platform == "win32" or sys.platform == "cygwin":
                # change readchar key UP & DOWN by 'w' and 's'
                readchar.key.UP = 'w'
                readchar.key.DOWN = 's'
                arrow_alias_msg = " 'w' to Up, 's' to Down,"

            projects = [
                (f"{p.get('workspace_name')}/{p.get('name')}" if p.get('workspace_name') else p.get('name'),
                 PipeRiderProject(p))
                for p in project_list]

            if len(projects) == 1:
                _, p = projects[0]
                return p

            question = [
                inquirer.List('selected_project',
                              message=f"Please select a project as default project ({arrow_alias_msg} ENTER to confirm )",
                              choices=projects,
                              carousel=True,
                              )
            ]
            answers = inquirer.prompt(question, raise_keyboard_interrupt=True)
            if answers:
                return answers['selected_project']
            return None

        def _select_project_by_rich(project_list: list) -> PipeRiderProject:
            console = Console()
            console.print('[[yellow]?[/yellow]] Please select a project as default project:')
            i = 0
            for p in project_list:
                i += 1
                if p.get('workspace_name'):
                    project_name = f"{p.get('workspace_name')}/{p.get('name')}"
                else:
                    project_name = p.get('name')
                console.print(f"[[blue]{i}[/blue]] {project_name}")

            while True:
                try:
                    type_idx = Prompt.ask('[[yellow]?[/yellow]] Select a number')
                    type_idx = int(type_idx)
                except EOFError as e:
                    raise e
                except Exception:
                    type_idx = 0
                if type_idx > len(project_list) or type_idx < 1:
                    console.print('    [[red]Error[/red]] Input is not a valid index value. Please try again.')
                else:
                    answer = PipeRiderProject(project_list[type_idx - 1])
                    break
            return answer if answer is not None else ''

        if project_name is None:
            existing_project_list = piperider_cloud.list_projects()
            if no_interaction is True:
                if len(existing_project_list) == 0:
                    console.print('[[bold red]Error[/bold red]] No project exists')
                    return 1
                project = PipeRiderProject(existing_project_list[0])
            elif FANCY_USER_INPUT:
                project = _select_project_by_inquirer(existing_project_list)
            else:
                project = _select_project_by_rich(existing_project_list)
        else:
            project = piperider_cloud.get_project_by_name(project_name)
            if project is None:
                console.print(f'[[bold red]Warning[/bold red]] Project \'{project_name}\' does not exist')
                return 1

        name = project.name if project.workspace_name is None else f"{project.workspace_name}/{project.name}"
        piperider_cloud.set_default_project(name)

        # TODO: Add project name into the datasource config if datasource is not None
        console.print(f'[[bold green]Config[/bold green]] Default project is set to \'{name}\'')
        return 0



import json
from typing import Callable, Dict

import pytest

from piperider_cli.dbt import dbt_version
from piperider_cli.dbt.reverse_manifest import reverse_to_run


@pytest.mark.skipif(
    dbt_version < "v1.5", reason="skip manifest test before dbt-core 1.5"
)
def test_reversed_manifest(
    sc_31711_run_data: Dict, piperider_schema_validate_func: Callable[[Dict], None]
):
    run_data: Dict = sc_31711_run_data
    manifest_data = run_data.get("dbt", {}).get("manifest")

    fake_run = reverse_to_run(manifest_data)
    piperider_schema_validate_func(fake_run)

    assert manifest_data == fake_run.get("dbt", {}).get("manifest")
    with open("sample.json", "w") as fh:
        fh.write(json.dumps(fake_run))

    # it is impossible to generate the same tables information without profiling
    # assert fake_run.get('tables') == run_data.get('tables')



import json
import os
import shutil
import sys
from datetime import datetime
from typing import Dict, List, Optional

import inquirer
import readchar
from rich.console import Console

import piperider_cli.hack.inquirer as inquirer_hack
from piperider_cli import clone_directory, datetime_to_str, open_report_in_browser, \
    raise_exception_when_directory_not_writable, str_to_datetime
from piperider_cli.configuration import Configuration, ReportDirectory
from piperider_cli.dbt.changeset import SummaryChangeSet
from piperider_cli.dbt.utils import ChangeType
from piperider_cli.generate_report import setup_report_variables
from piperider_cli.githubutil import fetch_pr_metadata
from piperider_cli.utils import create_link, remove_link


class RunOutput(object):
    def __init__(self, path):
        self.path = path
        self.name = None
        self.created_at = None

        self.table_count = 0
        self.pass_count = 0
        self.fail_count = 0
        self.cloud = None
        self.file_size = os.path.getsize(path)

        try:
            with open(path, 'r') as f:
                run_result = json.load(f)
                self.name = run_result['datasource']['name']
                self.created_at = run_result['created_at']

                tables = run_result.get('tables', {})
                self.table_count = len(tables.keys())

                for test in run_result.get('tests', []):
                    if test.get('status') == 'passed':
                        self.pass_count += 1
                    else:
                        self.fail_count += 1
                self.cloud = run_result.get('cloud')
        except Exception as e:
            if isinstance(e, json.decoder.JSONDecodeError):
                raise json.decoder.JSONDecodeError(
                    f'Invalid JSON in file "{path}"', e.doc, e.pos)
            raise e

    def verify(self) -> bool:
        # TODO: add some verification logic
        return True

    def load(self):
        with open(self.path, 'r') as f:
            data = json.load(f)
        return data

    def refresh(self):
        self.__init__(self.path)

    def __str__(self):
        created_at_str = datetime_to_str(str_to_datetime(self.created_at),
                                         to_tzlocal=True)

        def human_readable_size(size, decimal_places=2):
            for unit in ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB']:
                if size < 1024.0 or unit == 'PiB':
                    break
                size /= 1024.0
            return f"{size:.{decimal_places}f} {unit}"

        return f'{self.name:12} ' \
               f'#table={self.table_count:<6} ' \
               f'#pass={self.pass_count:<5} ' \
               f'#fail={self.fail_count:<5} ' \
               f'size={human_readable_size(self.file_size):<12}' \
               f'{created_at_str}'


def _merge_keys(base: List[str], target: List[str]):
    '''
    Merge keys from base, target tables. Unlike default union, it preserves the order for column rename, added, removed.

    :param base: keys for base table
    :param target: keys for base table
    :return: merged keys
    '''

    result = []
    while base and target:
        if base[0] == target[0]:
            result.append(base[0])
            base.pop(0)
            target.pop(0)
        elif base[0] in target:
            idx = target.index(base[0])
            for i in target[0:idx]:
                if i not in result:
                    result.append(i)
            result.append(base[0])
            base.pop(0)
            target = target[idx + 1:]
        else:
            result.append(base[0])
            base.pop(0)

    for c in base:
        if c not in result:
            result.append(c)

    for c in target:
        if c not in result:
            result.append(c)

    return result


def join(base, target):
    '''
    Join base and target to a dict which

    keys = (base keys) +  (target keys)
    result[keys] = {base: {...}, target: {...}

    :param base:
    :param target:
    :return:
    '''
    if not base:
        base = dict()
    if not target:
        target = dict()

    keys = _merge_keys(list(base.keys()), list(target.keys()))
    result = dict()
    for key in keys:
        value = dict()
        value['base'] = base.get(key)
        value['target'] = target.get(key)
        result[key] = value
    return result


class ComparisonData(object):
    STATE_ADD = 0
    STATE_DEL = 1
    STATE_MOD = 2

    def __init__(self, base: Dict, target: Dict, tables_from):
        self._id = datetime.now().strftime("%Y%m%d%H%M%S")

        if tables_from == 'target-only':
            target_run_tables = target.get('tables', {}).keys()
            target_run_metrics = {metric.get('name') for metric in target.get('metrics', [])}
            target_run_tests = {test.get('id') for test in target.get('tests', [])}

            base['tables'] = {table_name: base['tables'][table_name] for table_name in base['tables'] if
                              table_name in target_run_tables}
            if 'metrics' in base:
                base['metrics'] = [metric for metric in base['metrics'] if metric.get('name') in target_run_metrics]

            if 'tests' in base:
                base['tests'] = [test for test in base['tests'] if test.get('id') in target_run_tests]
        elif tables_from == 'base-only':
            base_run_tables = base.get('tables', {}).keys()
            base_run_metrics = {metric.get('name') for metric in base.get('metrics', [])}
            base_run_tests = {test.get('id') for test in base.get('tests', [])}

            target['tables'] = {table_name: target['tables'][table_name] for table_name in target['tables'] if
                                table_name in base_run_tables}
            if 'metrics' in target:
                target['metrics'] = [metric for metric in target['metrics'] if metric.get('name') in base_run_metrics]

            if 'tests' in target:
                target['tests'] = [test for test in target['tests'] if test.get('id') in base_run_tests]

        self._base = base
        self._target = target

        self.implicit = []
        self.explicit = []
        self.metadata = {}

        self.summary_change_set: Optional[SummaryChangeSet] = None
        self._update_implicit_and_explicit_changeset()
        self._update_github_pr_info()

    def _update_implicit_and_explicit_changeset(self):
        try:
            from piperider_cli.dbt.changeset import GraphDataChangeSet
            c = GraphDataChangeSet(self._base, self._target)
            self.explicit = c.list_explicit_changes()
            self.implicit = c.list_implicit_changes()

            self.summary_change_set = SummaryChangeSet(self._base, self._target)
        except BaseException as e:
            self.warning_for_legacy_metrics(e)

            console = Console()
            console.print('[bold yellow]Warning:[/bold yellow]')
            if isinstance(e, ValueError) and e.args:
                for line in e.args[0]:
                    console.print(f'  {line}', end='\n')
            else:
                console.print(e)
            console.print('\nGot problem to generate changeset.')

    def warning_for_legacy_metrics(self, e):
        if not hasattr(e, 'MESSAGE') or getattr(e, 'MESSAGE') != 'Compilation Error':
            return

        """"
        Example for "msg":
        'model.git_repo_analytics.commit_weekly' depends on 'metric.git_repo_analytics.total_commits' which is not in the graph!
        """
        if not hasattr(e, 'msg') or "depends on" not in getattr(e, 'msg'):
            return

        from piperider_cli.error import PipeRiderError
        raise PipeRiderError("Found legacy metrics",
                             hint="The dbt_metrics package has been deprecated and replaced with MetricFlow.")

    def _update_github_pr_info(self):
        if os.environ.get('GITHUB_EVENT_PATH'):
            self.metadata = fetch_pr_metadata()

    def id(self):
        return self._id

    def to_json(self):
        output = dict(
            created_at=datetime_to_str(datetime.utcnow()),
            base=self._base,
            # TODO: rename input -> target in schema and result json
            input=self._target,
            implicit=self.implicit,
            explicit=self.explicit,
            metadata=self.metadata,
        )
        return json.dumps(output, separators=(',', ':'))

    def to_summary_markdown_ng(self):
        if self._base.get('dbt') is None or self._target.get('dbt') is None:
            console = Console()
            console.print("[bold yellow]Warning: [/bold yellow]'summary.md' report is not generated.")
            console.print("To generate a summary.md file, please run the 'piperider run' command in a dbt project "
                          "and use the latest version of piperider.")
            return ""

        # TODO replace to new generator
        if self.summary_change_set:
            return self.summary_change_set.generate_markdown()

        return ""

    def to_cli_stats(self, console):
        console.print()

        if self.summary_change_set is None:
            return

        console.print("Impact Summary:")

        change_set = self.summary_change_set.models.explicit_changeset
        change_set += self.summary_change_set.metrics.explicit_changeset
        change_set += self.summary_change_set.seeds.explicit_changeset
        added = [x for x in change_set if x.change_type == ChangeType.ADDED]
        removed = [x for x in change_set if x.change_type == ChangeType.REMOVED]
        modified = [x for x in change_set if x.change_type == ChangeType.MODIFIED]
        code_change = [
            f"added={len(added)}",
            f"removed={len(removed)}",
            f"modified={len(modified)}",
        ]

        potentially_impacted = [x.unique_id for x in
                                (self.summary_change_set.models.modified_with_downstream + removed)]
        impacted = [x for x in list(set(
            self.summary_change_set.models.diffs + self.summary_change_set.metrics.diffs + self.summary_change_set.seeds.diffs
        )) if x in potentially_impacted]
        assessed_no_impacted = [x for x in list(set(
            self.summary_change_set.models.no_diffs + self.summary_change_set.metrics.no_diffs + self.summary_change_set.seeds.no_diffs
        )) if x in potentially_impacted]
        impact_summary = [
            f"potentially_impacted={len(potentially_impacted)}",
            f"assessed={len(impacted) + len(assessed_no_impacted)}",
            f"skipped={len(potentially_impacted) - len(impacted) - len(assessed_no_impacted)}",
            f"impacted={len(impacted)}",
        ]

        console.print("  Code Changes: " + ", ".join(code_change))
        console.print("  Resource Impact: " + ", ".join(impact_summary))

        console.print("")


def prepare_default_output_path(filesystem: ReportDirectory, created_at):
    latest_symlink_path = os.path.join(filesystem.get_comparison_dir(), 'latest')
    comparison_path = os.path.join(filesystem.get_comparison_dir(), created_at)

    if not os.path.exists(comparison_path):
        os.makedirs(comparison_path, exist_ok=True)

    # Create a symlink pointing to the latest comparison directory
    remove_link(latest_symlink_path)

    console = Console()
    if not os.path.exists(latest_symlink_path):
        create_link(comparison_path, latest_symlink_path)
    else:
        console.print(f'[bold yellow]Warning: {latest_symlink_path} already exists[/bold yellow]')

    return comparison_path


class CompareReport(object):
    def __init__(self, profiler_output_path, a=None, b=None, datasource=None, profiler_outputs=None):
        self.profiler_output_path = profiler_output_path
        self.profiler_outputs = profiler_outputs
        self.console = Console()
        self.a: RunOutput = RunOutput(a) if a else None
        self.b: RunOutput = RunOutput(b) if b else None
        self.datasource = datasource

    def list_existing_outputs(self, output_search_path=None) -> List[RunOutput]:
        """
        List existing profiler outputs.
        """

        def _walk_throw_runs(path):
            outputs = []
            for root, dirs, _ in os.walk(path):
                for dir_name in dirs:
                    if dir_name == 'latest':
                        continue
                    run_json = os.path.join(root, dir_name, 'run.json')
                    if not os.path.exists(run_json):
                        continue
                    output = RunOutput(run_json)
                    if self.datasource and output.name != self.datasource:
                        continue
                    outputs.append(output)
            outputs.sort(key=lambda x: (x.name, x.created_at), reverse=True)
            return outputs

        if output_search_path is None:
            output_search_path = self.profiler_output_path

        return _walk_throw_runs(output_search_path)

    def get_the_last_two_reports(self):
        outputs = self.list_existing_outputs()
        outputs.sort(key=lambda x: x.created_at)
        outputs = list(filter(lambda x: x.name == outputs[-1].name, outputs))
        if len(outputs) < 2:
            return None, None
        return outputs[-2:]

    def select_reports(self, use_last_two=None, reverse=False):
        if use_last_two:
            first, second = self.get_the_last_two_reports()
            if reverse:
                self.a, self.b = second, first
            else:
                self.a, self.b = first, second
        elif self.a is None and self.b is None:
            self.a, self.b = self.select_two_reports()
        elif self.a and self.b is None:
            self.b = self.select_one_report()
            pass
        elif self.a is None and self.b:
            self.a = self.select_one_report()

        if self.a and self.b:
            self.console.print('Selected reports:')
            self.console.print(f'  Base:   {self.a.path}', soft_wrap=True)
            self.console.print(f'  Target: {self.b.path}', soft_wrap=True)
            return True
        return False

    def select_multiple_reports(self, action='compare', limit=None):
        def _report_validater(answers, current) -> bool:
            if limit is None:
                return len(current) > 0
            else:
                return len(current) == limit

        profiler_outputs = self.profiler_outputs if self.profiler_outputs is not None else self.list_existing_outputs()
        arrow_alias_msg = ''
        if sys.platform == "win32" or sys.platform == "cygwin":
            # change readchar key UP & DOWN by 'w' and 's'
            readchar.key.UP = 'w'
            readchar.key.DOWN = 's'
            arrow_alias_msg = " 'w' to Up, 's' to Down,"

        if limit is None:
            if len(profiler_outputs) == 0:
                raise Exception("Not enough reports to compare. Please run 'piperider run' first.")
            questions = [
                inquirer.Checkbox('profiler_output',
                                  message=f"Please select the reports to {action} ({arrow_alias_msg} SPACE to select, and ENTER to confirm )",
                                  choices=profiler_outputs,
                                  carousel=True,
                                  validate=_report_validater,
                                  )
            ]
        else:
            if len(profiler_outputs) < limit:
                raise Exception("Not enough reports to compare. Please run 'piperider run' first.")
            report_msg = 'a report' if limit == 1 else f'the {limit} reports'
            questions = [
                inquirer_hack.LimitedCheckboxQuestion('profiler_output',
                                                      message=f"Please select {report_msg} to {action} ({arrow_alias_msg} SPACE to select, and ENTER to confirm )",
                                                      choices=profiler_outputs,
                                                      carousel=True,
                                                      validate=_report_validater,
                                                      limited=limit,
                                                      )
            ]

        answers = inquirer_hack.prompt_ex(questions, raise_keyboard_interrupt=True)
        if answers:
            return answers['profiler_output']
        return None

    def select_one_report(self, action='compare') -> RunOutput:
        answers = self.select_multiple_reports(action=action, limit=1)
        if answers:
            return answers[0]
        else:
            return None

    def select_two_reports(self, action='compare') -> (RunOutput, RunOutput):
        """
        Select multiple files from a list of files.
        """

        answers = self.select_multiple_reports(action=action, limit=2)
        if answers:
            return answers[0], answers[1]
        return None, None

    def generate_data(self, tables_from: str) -> ComparisonData:
        if self.a is None or self.b is None:
            raise Exception("Please select reports to compare first.")

        return ComparisonData(self.a.load(), self.b.load(), tables_from)

    @staticmethod
    def exec(*, a=None, b=None, last=None, datasource=None, report_dir=None, output=None, tables_from='all',
             summary_file=None, force_upload=False, enable_share=False, open_report=False, project_name: str = None,
             debug=False, show_progress=False, reverse_last=False):
        console = Console()
        console.rule('Comparison report', style='bold blue')

        filesystem = Configuration.instance().activate_report_directory(report_dir=report_dir)
        raise_exception_when_directory_not_writable(output)

        report = CompareReport(filesystem.get_output_dir(), a, b, datasource=datasource)
        if not report.select_reports(use_last_two=last, reverse=reverse_last):
            raise Exception('No valid reports found')

        from piperider_cli.cloud_connector import CloudConnector
        report_url = None
        summary_data = None

        if force_upload and (report.a.cloud is None or report.a.cloud.get('project_name') != project_name):
            CloudConnector.upload_report(report.a.path, show_progress=show_progress, project_name=project_name)
            report.a.refresh()

        if force_upload and (report.b.cloud is None or report.b.cloud.get('project_name') != project_name):
            CloudConnector.upload_report(report.b.path, show_progress=show_progress, project_name=project_name)
            report.b.refresh()

        # Generate comparison report URL & summary markdown
        if report.a.cloud and report.b.cloud:
            base = str(report.a.cloud.get('run_id'))
            target = str(report.b.cloud.get('run_id'))
            project_name = report.a.cloud.get('project_name')
            response = CloudConnector.generate_compare_report(base, target, project_name=project_name, debug=False)
            if response:
                report_url = response.get('url')
                summary_data = response.get('summary')

        comparison_data = report.generate_data(tables_from)

        from piperider_cli import data
        report_template_dir = os.path.join(os.path.dirname(data.__file__), 'report', 'comparison-report')
        with open(os.path.join(report_template_dir, 'index.html')) as f:
            report_template_html = f.read()

        def output_report(directory):
            clone_directory(report_template_dir, directory)
            filename = os.path.join(directory, 'index.html')
            with open(filename, 'w', encoding='utf-8') as f:
                html = setup_report_variables(report_template_html, False, comparison_data.to_json())
                f.write(html)

        def output_summary(directory, summary_data):
            filename = os.path.join(directory, 'summary.md')
            with open(filename, 'w', encoding='utf-8') as f:
                f.write(summary_data)

        data_id = comparison_data.id()
        summary_data = summary_data if summary_data else comparison_data.to_summary_markdown_ng()

        default_report_directory = prepare_default_output_path(filesystem, data_id)
        output_report(default_report_directory)
        if summary_data:
            output_summary(default_report_directory, summary_data)

        comparison_dir = filesystem.get_comparison_dir()
        report_path = os.path.join(comparison_dir, 'latest', 'index.html')
        summary_md_path = os.path.join(comparison_dir, 'latest', 'summary.md')
        if not os.path.exists(os.path.join(comparison_dir, 'latest')):
            report_path = os.path.join(default_report_directory, 'index.html')
            summary_md_path = os.path.join(default_report_directory, 'summary.md')

        if enable_share:
            if report.a.cloud is None or report.b.cloud is None:
                console.print(
                    '[[bold yellow]Skip[/bold yellow]] Please enable cloud auto upload or use "piperider compare --upload" to upload reports to cloud first.')
            else:
                from piperider_cli.cloud_connector import CloudConnector
                base = str(report.a.cloud.get('run_id'))
                target = str(report.b.cloud.get('run_id'))
                CloudConnector.share_compare_report(base, target, project_name=project_name)

        if output:
            clone_directory(default_report_directory, output)
            report_path = os.path.abspath(os.path.join(output, 'index.html'))
            summary_md_path = os.path.abspath(os.path.join(output, 'summary.md'))

        if summary_file:
            summary_file = os.path.abspath(summary_file)
            summary_dir = os.path.dirname(summary_file)
            if summary_dir:
                os.makedirs(summary_dir, exist_ok=True)
            shutil.copyfile(summary_md_path, summary_file)
            summary_md_path = summary_file

        comparison_data.to_cli_stats(console)
        console.print()
        console.print(f"Comparison report: {report_path}")
        if summary_data:
            console.print(f"Comparison summary: {summary_md_path}")
        if report_url:
            console.print(f"Comparison report URL: {report_url}?utm_source=cli", soft_wrap=True)

        if open_report:
            if report_url:
                open_report_in_browser(report_url, True)
            else:
                open_report_in_browser(report_path)

        if debug:
            # Write comparison data to file
            with open(os.path.join(default_report_directory, 'comparison_data.json'), 'w', encoding='utf-8') as f:
                f.write(comparison_data.to_json())



import math
from typing import Optional

from .recommender_assertion import RecommendedAssertion


def recommended_table_row_count_assertion(table, column, profiling_result) -> RecommendedAssertion:
    if column is not None:
        return None

    row_count = profiling_result['tables'][table].get('row_count')
    if row_count is None:
        return None

    test_metric_name = 'row_count'
    assertion_values = {
        'gte': int(row_count * 0.9),
    }
    assertion = RecommendedAssertion(None, test_metric_name, assertion_values)
    return assertion


def recommended_column_schema_type_assertion(table, column, profiling_result) -> RecommendedAssertion:
    if column is None:
        return None

    schema_type = profiling_result['tables'][table]['columns'][column].get('schema_type')
    if schema_type is None:
        return None

    test_function_name = 'assert_column_schema_type'
    assertion_values = {
        'schema_type': schema_type
    }
    assertion = RecommendedAssertion(test_function_name, None, assertion_values)
    return assertion


def recommended_column_min_assertion(table, column, profiling_result) -> RecommendedAssertion:
    if column is None:
        return None

    column_metric = profiling_result['tables'][table]['columns'][column]
    column_type = column_metric['type']
    if column_type == 'numeric':
        valids = column_metric.get('valids')
        if not valids:
            return None

        column_min = column_metric['min']
        histogram_counts = column_metric['histogram']['counts'] if column_metric['histogram'] else []

        count = 0
        for i, v in enumerate(reversed(histogram_counts)):
            count = count + v
            if i == len(histogram_counts) // 2:
                break

        if count / valids > 0.95:
            test_function_name = 'assert_column_min_in_range'
            assertion_values = {
                'min': sorted([round(column_min * 0.9, 4), round(column_min * 1.1, 4)])
            }
            assertion = RecommendedAssertion(test_function_name, None, assertion_values)
            return assertion
    else:
        return None


def recommended_column_max_assertion(table, column, profiling_result) -> RecommendedAssertion:
    if column is None:
        return None

    column_metric = profiling_result['tables'][table]['columns'][column]
    column_type = column_metric['type']
    if column_type == 'numeric':
        valids = column_metric.get('valids')
        if not valids:
            return None

        column_max = column_metric['max']
        histogram_counts = column_metric['histogram']['counts'] if column_metric['histogram'] else []

        count = 0
        for i, v in enumerate(histogram_counts):
            count = count + v
            if i == len(histogram_counts) // 2:
                break

        if count / valids > 0.95:
            test_function_name = 'assert_column_max_in_range'
            assertion_values = {
                'max': sorted([round(column_max * 0.9, 4), round(column_max * 1.1, 4)])
            }
            assertion = RecommendedAssertion(test_function_name, None, assertion_values)
            return assertion
    else:
        return None


def recommended_column_value_assertion(table, column, profiling_result) -> Optional[RecommendedAssertion]:
    if column is None:
        return None

    column_metric = profiling_result['tables'][table]['columns'][column]
    column_type = column_metric['type']
    assert_name = 'assert_column_value'
    if column_type == 'numeric':
        valids = column_metric.get('valids')
        if not valids:
            return None

        histogram_counts = column_metric['histogram']['counts'] if column_metric['histogram'] else []
        if not histogram_counts:
            return None

        column_min = column_metric.get('min')
        column_max = column_metric.get('max')

        if column_min is None or column_max is None:
            return None

        count_first_half = sum(histogram_counts[0:len(histogram_counts) // 2])
        count_second_half = sum(histogram_counts[math.ceil(len(histogram_counts) / 2):])

        if count_first_half / valids > 0.95:
            boundary = column_max * 1.1 if column_max > 0 else column_max * 0.9
            assertion_values = {
                'lte': round(boundary, 4)
            }
            return RecommendedAssertion(assert_name, None, assertion_values)
        elif count_second_half / valids > 0.95:
            boundary = column_min * 0.9 if column_min > 0 else column_min * 1.1
            assertion_values = {
                'gte': round(boundary, 4)
            }
            return RecommendedAssertion(assert_name, None, assertion_values)
    elif column_type == 'string':
        distinct = column_metric.get('distinct')
        if distinct is None:
            return None

        topk = []
        if column_metric.get('topk'):
            topk = column_metric.get('topk').get('values', [])

        if distinct and distinct <= 50 and topk:
            assertion_values = {
                'in': topk
            }
            return RecommendedAssertion(assert_name, None, assertion_values)
    else:
        return None


def recommended_column_unique_assertion(table, column, profiling_result) -> RecommendedAssertion:
    if column is None:
        return None

    column_metric = profiling_result['tables'][table]['columns'][column]
    column_type = column_metric.get('type')
    if column_type is None:
        return None

    if column_type == 'string':
        valids = column_metric.get('valids')
        distinct = column_metric.get('distinct')

        if valids is None or distinct is None:
            return None

        if valids > 0 and distinct == valids:
            test_function_name = 'assert_column_unique'
            assertion = RecommendedAssertion(test_function_name, None, None)
            return assertion
    else:
        return None


def recommended_column_not_null_assertion(table, column, profiling_result) -> RecommendedAssertion:
    if column is None:
        return None

    column_metric = profiling_result['tables'][table]['columns'][column]
    non_nulls = column_metric.get('non_nulls')
    if non_nulls is None:
        return

    total = column_metric['total']

    if total > 0 and non_nulls == total:
        test_function_name = 'assert_column_not_null'
        assertion = RecommendedAssertion(test_function_name, None, None)
        return assertion
    else:
        return None


RecommendedRules = [
    recommended_table_row_count_assertion,
    recommended_column_schema_type_assertion,
    recommended_column_unique_assertion,
    recommended_column_not_null_assertion,
    recommended_column_value_assertion
]



def disable_dbt_compile_stats():
    import dbt.compilation as compilation

    class context_class:

        def __enter__(self):
            self.original_print_compile_stats = compilation.print_compile_stats
            compilation.print_compile_stats = lambda x: None

        def __exit__(self, exc_type, exc_val, exc_tb):
            compilation.print_compile_stats = self.original_print_compile_stats

    try:
        from dbt.events.functions import cleanup_event_logger
        cleanup_event_logger()
    except ImportError:
        # we dont support the dbt version less than 1.4
        pass

    return context_class()


class DbtVersionTool:

    def __init__(self):
        from dbt import version as dbt_version
        self.dbt_version = self.parse(dbt_version.__version__)

    @staticmethod
    def parse(version: str):
        from packaging import version as v
        return v.parse(version)

    def as_version(self, other):
        from packaging.version import Version
        if isinstance(other, Version):
            return other
        if isinstance(other, str):
            return self.parse(other)
        return self.parse(str(other))

    def __ge__(self, other):
        return self.dbt_version >= self.as_version(other)

    def __gt__(self, other):
        return self.dbt_version > self.as_version(other)

    def __lt__(self, other):
        return self.dbt_version < self.as_version(other)

    def __le__(self, other):
        return self.dbt_version <= self.as_version(other)

    def __eq__(self, other):
        return self.dbt_version.release[:2] == self.as_version(other).release[:2]

    def __str__(self):
        return ".".join([str(x) for x in list(self.dbt_version.release)])


dbt_version = DbtVersionTool()



from . import DataSource
from .field import TextField
from ..error import PipeRiderConnectorError


class DataBricksDataSource(DataSource):
    def __init__(self, name, **kwargs):
        super().__init__(name, 'databricks', **kwargs)
        self.fields = [
            TextField('host', description='Databricks host'),
            TextField('token', description='Databricks token'),
            TextField('catalog', description='Databricks catalog name', optional=True),
            TextField('http_path', description='Databricks http path'),
            TextField('schema', description='Databricks schema name'),
        ]

    def validate(self):
        if self.type_name != 'databricks':
            raise ValueError('type name should be databricks')
        return self._validate_required_fields()

    def verify_connector(self):
        try:
            import databricks
        except Exception as e:
            return PipeRiderConnectorError(str(e), 'databricks')

    def to_database_url(self, database):
        credential = self.credential
        host = credential.get('host')
        token = credential.get('token')
        schema = credential.get('schema')
        return f'databricks://token:{token}@{host}:443/{schema}'
        pass

    def engine_args(self):
        credential = self.credential
        http_path = credential.get('http_path')
        args = {'http_path': http_path}

        if credential.get('catalog'):
            args['catalog'] = credential.get('catalog')

        return dict(connect_args=args)

    def _get_display_description(self):
        cred = self.credential
        return f"type={self.type_name}, database={cred.get('catalog')}, schema={cred.get('schema')}"

    def get_database(self):
        return self.credential.get('catalog')

    def get_schema(self):
        return self.credential.get('schema')
